-----------------------------------------------------------ASSETS LIST------------------------------------------------------------------
Assets =
{
    Asset("IMAGE", "images/back.tex"),
    Asset("ATLAS", "images/back.xml"),
    Asset("IMAGE", "images/neck.tex"),
    Asset("ATLAS", "images/neck.xml"),
	Asset("ANIM", "anim/teleportato_parts.zip"),
    Asset("ANIM", "anim/teleportato_parts_build.zip"),
    Asset("ANIM", "anim/teleportato_adventure_parts_build.zip"),
    Asset("IMAGE", "images/inventoryimages/kokocollar.tex"),
    Asset("ATLAS", "images/inventoryimages/kokocollar.xml"),
	Asset("ANIM", "anim/nightmaretreat.zip"),
    Asset("IMAGE", "images/inventoryimages/nightmaretreat.tex"),
    Asset("ATLAS", "images/inventoryimages/nightmaretreat.xml"),
}

PrefabFiles = 
{
    "telebase1",
	"bell",
	"dug_reeds",
	"critters",
	"koko",
    "kokocollar",
    "nightmaretreat",
    "lana",
    "denmother",
    "denmothercollar",
    "denguard",
}
-------------------------------------------------WORLD CLEANING - RYUU----------------------------------------------
local require = GLOBAL.require

modimport("PrefabsToClean.lua")

-- Time in seconds
local CLEAN_WARNING_TIME = GetModConfigData("CLEAN_WARNING_TIME")
local CLEAN_TIME = GetModConfigData("CLEAN_TIME")

CLEAN_WARNING_TIME = (CLEAN_TIME - CLEAN_WARNING_TIME) > 1 and CLEAN_WARNING_TIME or 5

local PrefabsToClean = GLOBAL.PREFABSTOCLEAN.Prefabs
local ExcludeTags = GLOBAL.PREFABSTOCLEAN.ExcludeTags
local BirdTreasure = GLOBAL.PREFABSTOCLEAN.BirdTreasure

local PrefabCount = {}

local PrefabsRemoved = 0

local function FormatN(n)
	return string.format("%01.f", n)
end

local function PluralS(n)
	return n > 1 and "s" or ""
end

local function FormatTime(t)
	if t == 0 then
		return "instantly"
	else
		local	hours = math.floor(t/3600)
		local	mins = math.floor(t/60 - hours*60)
		local	secs = math.floor(t - hours*3600 - mins *60)
		
		local suffix = secs > 0 and " second" .. PluralS(secs) or ""
		suffix = mins > 0 and " minute" .. PluralS(mins) or suffix
		suffix = hours > 0 and " hour" .. PluralS(hours) or suffix
		
		if hours == 0 and mins == 0 then
			return secs .. suffix
		elseif hours == 0 and secs == 0 then
			return mins .. suffix
		elseif hours == 0 then
			return mins .. ":" .. secs .. suffix
		elseif mins == 0 then
			-- return hours .. ":" .. mins .. ":" .. secs .. " hours" -- ignore seconds, they don't matter at this point
			return hours .. suffix
		else
			return hours .. ":" .. mins .. suffix
		end
	end
	
	return "???"
end

local function AnnounceMsg(msg)
    GLOBAL.TheNet:Announce(msg, nil, nil, nil)
end

local function IsInInventory(inst)
    if inst.components.inventoryitem ~= nil and inst.components.inventoryitem.owner ~= nil then
        return true
    end
	
    return false
end

local function InitPrefabCountTable()
	for k,v in pairs(PrefabsToClean) do
		if PrefabsToClean[k] > 0 then PrefabCount[k] = 0 end
	end
end

local function IsTaggedExclude(inst)
	for i,v in ipairs(ExcludeTags[inst.prefab]) do
		if inst:HasTag(v) then return true end
	end

	return false
end


local function CleanPrefabs()
	InitPrefabCountTable()

	for k,v in pairs(GLOBAL.Ents) do
		if PrefabsToClean[v.prefab] and not IsInInventory(v) and not (ExcludeTags[v.prefab] and IsTaggedExclude(v)) then
			if PrefabsToClean[v.prefab] == 0 or PrefabCount[v.prefab] > PrefabsToClean[v.prefab] then
				v:Remove()
				PrefabsRemoved = PrefabsRemoved + 1
			else
				PrefabCount[v.prefab] = PrefabCount[v.prefab] + 1
			end
		end
	end
end

function CleanWorld()
	local worldname = not GLOBAL.TheWorld:HasTag("cave") and "world" or "caves"
	
	AnnounceMsg("Cleaning the " .. worldname .. "...")
	
	PrefabsRemoved = 0
	CleanPrefabs()
	AnnounceMsg("Removed " .. PrefabsRemoved .. " objects from the " .. worldname .. "!")
	
	AnnounceMsg("Next clean will be in " .. FormatTime(CLEAN_TIME))
	
	GLOBAL.TheWorld.clean_task = GLOBAL.TheWorld:DoTaskInTime(CLEAN_TIME - CLEAN_WARNING_TIME, CleanWarning)
end

function CleanWarning()
	local worldname = not GLOBAL.TheWorld:HasTag("cave") and "World" or "Caves"
	
	AnnounceMsg("Warning! " .. worldname .. " will be cleaned in " .. FormatTime(CLEAN_WARNING_TIME) .. "!")
	
	GLOBAL.TheWorld.clean_task = GLOBAL.TheWorld:DoTaskInTime(CLEAN_WARNING_TIME, CleanWorld)
end


local function AddWorldPostInit(world)
	if not GLOBAL.TheWorld.ismastersim then return end
		
	world.clean_task = world:DoTaskInTime(CLEAN_TIME - CLEAN_WARNING_TIME, CleanWarning)
end


AddPrefabPostInit("forest", AddWorldPostInit)
AddPrefabPostInit("cave", AddWorldPostInit)

GLOBAL.c_cleanworld = function(do_warning)
	if not do_warning then
		CleanWorld()
	else
		CleanWarning()
	end
end

-------------------------- Bait items ----------------------------

local birds = {"robin", "robin_winter"}

GLOBAL.FOODTYPE.BIRDTREASURE = "BIRDTREASURE"

local function MakeBirdFood(inst)
	inst:AddTag("edible_BIRDTREASURE")
	if not inst.components.bait then
		inst:AddComponent("bait")
	end
	if not inst.components.edible then
		inst:AddComponent("edible")
		inst.components.edible.foodtype = GLOBAL.FOODTYPE.BIRDTREASURE
	end
end


local function MakeBirdHungry(inst)
	if inst.components.eater == nil then return end
	
	if inst.components.eater.preferseating then
		table.insert(inst.components.eater.preferseating, GLOBAL.FOODTYPE.BIRDTREASURE)
		table.insert(inst.components.eater.caneat, GLOBAL.FOODTYPE.BIRDTREASURE)
	else
		table.insert(inst.components.eater.foodprefs, GLOBAL.FOODTYPE.BIRDTREASURE)
	end
end


local function DoBirdEatItems()
	for i,v in ipairs(BirdTreasure) do
		AddPrefabPostInit(v, MakeBirdFood)
	end
	
	for i,v in ipairs(birds) do
		AddPrefabPostInit(v, MakeBirdHungry)
	end
end


if true then DoBirdEatItems() end

--[[
AddComponentPostInit("herd", function(self)
    local oldOnRemoveEntity = self.OnRemoveEntity
    function self.OnRemoveEntity() = function(self)
        for i,v in ipairs(self.members) do
            self:RemoveMember(v)
        end
        
        oldOnRemoveEntity(inst)
    end
end)]]--


------------------------------------------------------ISLANDS----------------------------------------------------------
_G=GLOBAL
local TheNet = _G.TheNet
local TheSim = _G.TheSim
local TheWorld = _G.TheWorld




function getOwner(inst)
    local owner = nil
    if inst.components.inventoryitem ~= nil and inst.components.inventoryitem.owner~= nil then 
        if inst.components.inventoryitem.owner:HasTag("player") then
            owner = inst.components.inventoryitem.owner
        end

        if inst.components.inventoryitem.owner.components.inventoryitem and inst.components.inventoryitem.owner.components.inventoryitem.owner
            and inst.components.inventoryitem.owner.components.inventoryitem.owner:HasTag("player") then
            owner = inst.components.inventoryitem.owner.components.inventoryitem.owner         
        end
    end
    return owner
end


function AddAssets(arr)
    for i,v in ipairs(arr) do
        table.insert(Assets,v)
    end
end


function AddPlayersPostInit(fn)
    for i,v in ipairs(_G.DST_CHARACTERLIST) do -- DST_CHARACTERLIST + ROG_CHARACTERLIST
        AddPrefabPostInit(v,fn)
    end
    for i,v in ipairs(_G.MODCHARACTERLIST) do
        AddPrefabPostInit(v,fn)
    end
end



local function checkOnAnyIsland(inst)
    print ("checking for any island")
    local x,y,z = inst:GetPosition():Get() 
    local isl = TheSim:FindEntities(x,y,z, 5, {"island_portal"})[1]
    return  isl ~= nil
end

local function checkOnIsland(inst, island)
    print ("checking island: "..island)
    local x,y,z = inst:GetPosition():Get() 
    local isl = TheSim:FindEntities(x,y,z, 20, {island})[1]
    return  isl ~= nil
end


local function superbell(inst)
    local TheWorld = _G.TheWorld
    inst:RemoveComponent("finiteuses")   
    inst:RemoveTag("molebait")

    if not GLOBAL.TheWorld.ismastersim then
        return inst
    end

    inst.components.instrument.onplayed = function (a, musician) 
    if (inst == nil) then print("empty inst inside inplayed") return true end       
    local iid = inst.islandid or "0"
    local Island = checkOnIsland(inst.components.inventoryitem.owner, "island"..iid)
    local IslandName = ""

--[[    if musician.components.sanity:GetPercent() < 50 then
        pickguy.components.talker:Say("One island is enough for me.")
    end]]

    if Island then 
        IslandName = "island0"
        --musician.components.sanity:DoDelta(1)
    else

        --[[if musician.components.sanity.current < 50 then
            musician.components.talker:Say("I am too crazy for teleporting...")
            return
        end]]

        IslandName = "island"..iid
        --musician.components.sanity:DoDelta(1)
    end

    --musician.components.hunger:DoDelta()

    print ("tp to "..IslandName)
    inst:DoTaskInTime(1, function(a) 
        local c = GLOBAL.SpawnPrefab("poopcloud")
        if a.components.inventoryitem.owner == nil then return end
        c.Transform:SetPosition(a.components.inventoryitem.owner:GetPosition():Get())

        --GLOBAL.TheWorld:PushEvent("ms_sendlightningstrike", a.components.inventoryitem.owner:GetPosition()) 

        local pos = GLOBAL.Vector3(a.Transform:GetWorldPosition())
        --PlayThunderSound(pos)
        a.SoundEmitter:PlaySound("dontstarve/rain/thunder_close")

        ---GLOBAL.TheWorld:PushEvent("ms_sendlightningstrike") 
        ---PlaySound("dontstarve/rain/thunder_close")
        musician:ShakeCamera(GLOBAL.CAMERASHAKE.FULL, .7, .02, .5, proxy, 40)



        inst:DoTaskInTime(0.5, function (a)
            local ents = GLOBAL.TheSim:FindEntities(0,0,0, 9001, {IslandName})
            local N = table.getn(ents)
            if (N==0) then return end
            local ent = ents[math.random(N)]

            if ent == nil then return end
            local x,y,z = ent.Transform:GetWorldPosition()
            --a.components.inventoryitem.owner
            musician.Transform:SetPosition(x, y, z)

            inst:DoTaskInTime(0.7, function(a) 
                if a.components.inventoryitem.owner == nil then return end
                local d = GLOBAL.SpawnPrefab("poopcloud")
                d.Transform:SetPosition(a.components.inventoryitem.owner:GetPosition():Get())
                --GLOBAL.TheWorld:PushEvent("ms_sendlightningstrike", a.components.inventoryitem.owner:GetPosition()) 
            end)
        end)
    end) 
end


    inst.components.instrument.onheard = function (a)  end   

   inst.OnSave = function(self, data)
--        local data = {}--origSave(self)
         
        data.islandid = self.islandid
    print("saving")
    print(self.islandid)
        --return data
     end


--local origLoad = inst.LoadPostPass
    inst.OnLoad = function(self, data)
    --origLoad(self,newents,savedata)
    if data then
            self.islandid = data.islandid

            if self.islandid then
        self:AddTag("bell_island"..self.islandid)
        end

        print("loading")
        print(self.islandid)
    end
     end
end




--Make the character talk sound
local function DoTalkSound(inst)
    if inst.talksoundoverride ~= nil then
        inst.SoundEmitter:PlaySound(inst.talksoundoverride, "talk")
        return true
    elseif not inst:HasTag("mime") then
        inst.SoundEmitter:PlaySound((inst.talker_path_override or "dontstarve/characters/")..(inst.soundsname or inst.prefab).."/talk_LP", "talk")
        return true
    end
end


local function OnPutInInventory(inst, pickguy)
    if inst.fixtask == nil then
        inst.fixtask = inst:DoTaskInTime(1, FixChester)
    end


    if pickguy.components.inventory ~= nil and pickguy.components.inventory:Has("bell", 1) then --We already have one
        if pickguy.components.inventory.activeitem ~= nil 
            and pickguy.components.inventory.activeitem.prefab == "bell" then        
            return
        end

        pickguy.components.inventory:DropEverythingWithTag("bell") --Drop current one
        if pickguy.components.talker then
            pickguy.components.talker:Say("One island is enough for me.") --Show text
            pickguy.SoundEmitter:KillSound("talk") --Stop talking if currently talking
            DoTalkSound(pickguy) --Add talk sound
            pickguy:DoTaskInTime(1.5, function(inst) inst.SoundEmitter:KillSound("talk") end) --Stop talking again after 1.5 seconds
        end
    end
end







AddPrefabPostInit("bell", function(inst) 
    superbell(inst)
    if not GLOBAL.TheWorld.ismastersim then
        return inst
    end
    inst.components.inventoryitem:SetOnPutInInventoryFn(OnPutInInventory)
    inst.AnimState:PushAnimation("idle")
    print("postinit") 
end
)



function checkPlace(x,y,r)
    local TheWorld = _G.TheWorld
    local res = true
    for tx=x-r,x+r do
        --local s= ""
        for ty=y-r,y+r do
            --local c = ""
            --c = TheWorld.Map:GetTile(tx, ty) 
            if TheWorld.Map:GetTile(tx, ty) ~= 1 then res=false end
            --s=s.."            "..c
        end
        --print(s)
    end
    return res
end

--ThePlayer.Transform:SetPosition(tx, 0, ty)


function GetTileCenter(x,y)
    local TheWorld = _G.TheWorld
    local x1, y1 = TheWorld.Map:GetTileCoordsAtPoint(0,0,0) 
    local x2, y2 = TheWorld.Map:GetTileCoordsAtPoint(1000,0,1000)   

    local xr = (x-x1)/(x2-x1)*1000
    local yr = (y-y1)/(y2-y1)*1000

    return xr,yr
end



function makeIsland(tx,ty, r)
    local TheWorld = _G.TheWorld
    for x = -r,r do 
        local s = ""
        for y = -r,r do
            local a="0"
            if (x*x+y*y<=r*r) then a = "*" end
            s=s..a

            if (x*x+y*y<=r*r) then 
                TheWorld.Map:SetTile(tx+x, ty+y, 7)
            end

        end
        --print(s)
    end


    local ux,uy = GetTileCenter(tx, ty)
    print(ux,uy)
    local base = GLOBAL.SpawnPrefab("telebase1")
    base.Transform:SetPosition(ux,0,uy)

    if TheWorld.N == nil then TheWorld.N = 0 end
    TheWorld.N = TheWorld.N + 1

    base.islandid = TheWorld.N
    print("island"..TheWorld.N)
    base:AddTag("island"..TheWorld.N)   

    local s = GLOBAL.SpawnPrefab("bell")
    s.islandid = TheWorld.N
    s:AddTag("bell_island"..TheWorld.N)
    print("bell_island"..TheWorld.N)
end


function makeIslands(r,N)
    local TheWorld = _G.TheWorld
    print("makeIslands "..r.." : "..N)
    for i=1,N do

        local s = TheWorld.Map:GetTileCoordsAtPoint(0,0,0)
        s=s*2

        local tx, ty = math.random(s), math.random(s)
        local q = 0
        while not checkPlace(tx,ty,r+10-q) do 
            if math.random(100)==1 then 
                    q=q+1 
                    if q>4 then q = 4 end
            end
            tx, ty = math.random(s), math.random(s)
        end 
        --local tx, ty = TheWorld.Map:GetTileCoordsAtPoint(ThePlayer:GetPosition():Get())
        --print(tx,ty,ThePlayer:GetPosition())
        --print(tx,ty,TheWorld.Map:GetTile(tx, ty))
        makeIsland(tx,ty, r)
    end
end


function SpreadPortals()
    local TheWorld = _G.TheWorld
    local nodes = TheWorld.topology.nodes
    for i, node in ipairs(nodes) do
        print(node.x, node.y)

        fr = GLOBAL.tonumber(GetModConfigData("Amount_of_ExitPortals"));

        if math.floor(i/fr)*fr==i then
            if (GLOBAL.RoadManager:IsOnRoad(node.x, 0, node.y)) then
                local base = GLOBAL.SpawnPrefab("telebase1")
                base.Transform:SetPosition(node.x,0,node.y)

                base.islandid = 0
                print("island0")
                base:AddTag("island0")          
            end
        end
    end
end


function RndBells(name)
    local TheSim = GLOBAL.TheSim
    local b = TheSim:FindEntities(0,0,0, 9001, {name})[1]
    if b == nil then
        return
    end
    b:RemoveTag("molebait")
    local x,y,z = b.Transform:GetWorldPosition()
    --print(x,y,z)
    --ThePlayer.Transform:SetPosition(x, y, z)

    local owner = nil
    if b.components.inventoryitem ~= nil and b.components.inventoryitem.owner~= nil then 
        if b.components.inventoryitem.owner:HasTag("player") then
            owner = b.components.inventoryitem.owner
        end

        if b.components.inventoryitem.owner.components.inventoryitem and b.components.inventoryitem.owner.components.inventoryitem.owner
            and b.components.inventoryitem.owner.components.inventoryitem.owner:HasTag("player") then

            owner = b.components.inventoryitem.owner.components.inventoryitem.owner         
        end
    end

    if owner == nil then
        print("free")
        

        if b.components.inventoryitem.owner ~= nil and b.components.inventoryitem.owner.components.inventory ~= nil then
            b.components.inventoryitem.owner.components.inventory:DropItem(b)
        end


        if b.components.inventoryitem.owner ~= nil and b.components.inventoryitem.owner.components.container~= nil then
            b.components.inventoryitem.owner.components.container:DropItem(b)
        end

        local ents = TheSim:FindEntities(0,0,0, 9001, {"island0"})
        local N = table.getn(ents)
        
        b:DoTaskInTime(math.random(10), function(b)

            if (N==0) then return end
            local ent = ents[math.random(N)]

            if ent == nil then return end
            local x,y,z = ent.Transform:GetWorldPosition()
    
                    b.Transform:SetPosition(x, y, z)
            print(b)
            print("to -> ")
            print(x,y,z)
        end)
        
    else
        print(owner)
    end
end

function c_announce(msg)
    local TheNet = _G.TheNet
    TheNet:Announce(msg, nil, nil, nil)
end





if TheNet:GetIsServer() then    
    AddPrefabPostInit("world",function (inst)
        inst:DoTaskInTime( 5, function()

            if not GLOBAL.TheWorld:HasTag("cave") then

                local isl = GLOBAL.TheSim:FindEntities(0,0,0, 9999, {"safe_exit_point"})[1]
                
                if isl ~= nil then  
                    c_announce("Teleporting bells to random portals...")
                    for i=1,60 do
                        RndBells("bell_island"..i)
                    end            
                end

                if isl == nil then            
                    SpreadPortals()    
                    makeIslands(10,GLOBAL.tonumber(GetModConfigData("Amount_of_10")))  --- Gigantic
                    makeIslands(9, GLOBAL.tonumber(GetModConfigData("Amount_of_9")))   --- huge
                    makeIslands(8, GLOBAL.tonumber(GetModConfigData("Amount_of_8")))   --- very large
                    makeIslands(7, GLOBAL.tonumber(GetModConfigData("Amount_of_7")))   --- large
                    makeIslands(6, GLOBAL.tonumber(GetModConfigData("Amount_of_6")))   --- very big
                    makeIslands(5, GLOBAL.tonumber(GetModConfigData("Amount_of_5")))   --- big
                    makeIslands(4, GLOBAL.tonumber(GetModConfigData("Amount_of_4")))   --- normal
                    makeIslands(3, GLOBAL.tonumber(GetModConfigData("Amount_of_3")))   --- small
                    makeIslands(2, GLOBAL.tonumber(GetModConfigData("Amount_of_2")))   --- tiny


                    --c_announce("Free bells were teleported to randomly")
                    c_announce("Islands were generated. World will reset")

                    inst:DoTaskInTime( 1, function()
                        GLOBAL.c_save()
                        inst:DoTaskInTime( 5, function()
                            GLOBAL.c_reset(true)
                        end)                
                    end)
                end
            end
			
        end)
    end)
end

if TheNet:GetIsServer() then    
    AddPrefabPostInit("world",function (inst)
        inst:DoTaskInTime( 5, function()

            if GLOBAL.TheWorld:HasTag("cave") then

                local isl = GLOBAL.TheSim:FindEntities(0,0,0, 9999, {"safe_exit_point"})[1]
                
                if isl ~= nil then  
                    c_announce("Teleporting bells to random portals...")
                    for i=1,21 do
                        RndBells("bell_island"..i)
                    end            
                end

                if isl == nil then            
                    SpreadPortals()    
                    makeIslands(15,GLOBAL.tonumber(GetModConfigData("Amount_of_15")))  --- Gigantic
                    makeIslands(14, GLOBAL.tonumber(GetModConfigData("Amount_of_14")))   --- huge
                    makeIslands(13, GLOBAL.tonumber(GetModConfigData("Amount_of_13")))   --- very large
                    makeIslands(12, GLOBAL.tonumber(GetModConfigData("Amount_of_12")))   --- large
                    makeIslands(11, GLOBAL.tonumber(GetModConfigData("Amount_of_11")))   --- very big




                    --c_announce("Free bells were teleported to randomly")
                    c_announce("Islands were generated. World will reset")

                    inst:DoTaskInTime( 1, function()
                        GLOBAL.c_save()
                        inst:DoTaskInTime( 5, function()
                            GLOBAL.c_reset(true)
                        end)                
                    end)
                end
            end
			
        end)
    end)
end




AddStategraphPostInit("wilson", function(sg)
    local old_death_fn = sg.events.death.fn
    sg.events.death.fn = function(inst, data)
        if checkOnAnyIsland(inst) then
            if inst.components.inventory ~= nil then
                c_announce("Death on an island happend.")
                inst.components.inventory:GiveItem(_G.SpawnPrefab("amulet"))                 
            end
        end

       return old_death_fn(inst, data)
    end
end)

GLOBAL.STRINGS.NAMES.BELL = "Island bell"
GLOBAL.STRINGS.CHARACTERS.GENERIC.DESCRIBE.BELL = "Seems like there is some sort of magic inside it..."
---------------------------------------------SHOW ME--------------------------------------------------------
local _G = GLOBAL

if modinfo then
	print("ShowMe version:",modinfo.version)
end

local GetGlobal=function(gname,default)
	local res=_G.rawget(_G,gname)
	if res == nil and default ~= nil then
		_G.rawset(_G,gname,default)
		return default
	end
	return res
end

--Locals (Add compatibility with any other mineable mods).
local mods = GetGlobal("mods",{})

local GetTime = _G.GetTime
local TheNet = _G.TheNet
local SERVER_SIDE = TheNet:GetIsServer()
local CLIENT_SIDE =	 TheNet:GetIsClient() or (SERVER_SIDE and not TheNet:IsDedicated())

local tonumber = _G.tonumber
local food_order = tonumber(GetModConfigData("food_order",true)) or 0
if food_order == 0 then
	food_order = tonumber(GetModConfigData("food_order")) or 0
end
--TUNING.SHOWME_FOOD_ORDER = food_order --для Tell Me
local food_style = tonumber(GetModConfigData("food_style",true)) or 0
if food_style == 0 then
	food_style = tonumber(GetModConfigData("food_style")) or 0
end
--TUNING.SHOWME_FOOD_STYLE = food_style

local food_estimation = tonumber(GetModConfigData("food_estimation",true)) or -1
if food_estimation == -1 then
	food_estimation = tonumber(GetModConfigData("food_estimation")) or 0
end

local show_food_units = tonumber(GetModConfigData("show_food_units",true)) or -1
if show_food_units == -1 then
	show_food_units = tonumber(GetModConfigData("show_food_units")) or 0
end

--Название на английском, краткая сетевая строка-алиас (для пересылки).
--Если алиас начинается с маленькой буквы, то он обязан быть длиной в 1 букву. Если с большой, то 2 (вторая буква может быть любого регистра).
local MY_STRINGS =
{ 
	{ armor = "Armor:" }, --A
	{ aggro = "Aggression:" }, --B
	{ cookpot = "Probably it will be" }, --C (Crock Pot)
	{ dmg = "Damage:" }, --D
	{ electric = "Electric:" }, --E --electric power
	{ food = "Food:"},
	{ S2="Now it's the summer," }, --G
	{ health= "Health:" }, --H --for food
	{ warm = "Insulation:" }, --I --winter insulation
	{ kill = "Kill:" }, --J	 --for Canibalism 18 mod
	{ kills = "Kills:" }, --K  --for Canibalism 18 mod
	{ loyal = "Loyality:" }, --L  --pigman and bunnyman
	{ S4="Now it's the autumn," }, --M
	{ remaining_days = "Remaining days" }, --N
	{ owner = "Owner:" }, --O --support of various mods
	{ perish = "Spoil in" }, --P -- Spoil in N days.
	{ hunger= "Hunger:" }, --Q
	{ range = "Range:" }, --R  --for range weapon or for mobs
	{ sanity= "Sanity:" }, --S
	{ thickness = "Thickness:" }, --T
	{ units_of = "units of" }, --U
	{ resist = "Resistance:" }, --V --against sleep darts, ice staff etc
	{ waterproof = "Waterproof:" }, --W
	{ heal = "Heal:" }, --X
	{ fishes = "Fishes:" }, --Y	 --in a pond
	{ fish = "Fish:" }, --Z
	{ sec= "Seconds:" },  --([)for cooking in Crock Pot
	{ love = "Love:" }, --(~)
	{ summer = "Summer:" }, --(])summer insulation
	{ absorb = "Absorb:" }, --(^)
	{ S3="Now it's the spring," }, --(_)
	{ is_admin = "This is the ADMIN!\nHe is out of the game\nso don't touch him." }, --(@)
	{ temperature = "t" }, --a
	{ hp= "HP:" }, --for characters
	{ armor_character = "Armor:" },
	{ sanity_character = "Sanity:" }, --S	
	{ fuel = "Fuel:" }, --F --for firepit
	{ speed = "Speed:" }, 
	{ uses_of = "uses of" },
	{ obedience = "Obedience:" },
	{ S1="Now it's the winter," }, 
	{ dmg_character = "Damage:" }, 
	{ power="Power:"},--usually means strengths of weapon but not physical damage
	{ cooldown="Cooldown:"},
	{ domest = "Tamed:" }, -- "Domestication:"
	{ will_die = "Will die in" }, --will die in N days (saying about pet or animal).
	{ will_dry = "Will dry in" },
	{ dmg_bonus = "Damage bonus:" }, -- Damage: +X (means damage modifier, not base damage)
	{ crop = "" }, --Not used. It's just a key for info type. Info - "Product: percent"
	{ grow_in = "Grow in" }, --About grass etc
	{ perish_product = "" }, --Just a key for info type. Info - "Product: time in days"
	{ just_time = "" }, --Just a key for info type. Info - [time]
	--Thirst mod
	{ water = "Water:" },
	{ salt = "Salt:" },
	{ sip = "A sip:" },
	{ watergainspeed = "Water gain speed:" },
	{ water_poisoned = "Is poisoned!" },
	
	{ timer = "Timer" },
	{ trade_gold = 'Gold tribute:' },
	{ trade_rock = 'Rock tribute:' },
	{ durability = 'Durability:' },
	{ strength = 'Strength:' },
	{ aoe = "AoE:" },
}

SHOWME_STRINGS = {
	loyal = "forever", --for very loyal pigman with loyalty over 9000
	of = " of ", -- X of Y (reserved)
	units_1 = "1 unit of ",
	units_many = " units of ",
	uses_1 = "1 use of ",
	uses_many = " uses of ", --X uses of Y, where X > 1
	days = " days", --Spoil in N days.
}

FOOD_TAGS = {}

MY_DATA = {}
for i,v in ipairs(MY_STRINGS) do
	for k,str in pairs(v) do --одна пара
		MY_DATA[k] = {
			desc = str,
			id = i,
			sym = string.char(i+64), -- A+
			fn = nil, --Function to return the proper string. By default: desc + " " + param1
			percent = nil, --To add "%" at the end of the number
		}
		v.key = k --Записываем ключ в отдельную переменную. Вообще это ужасный костыль, лучше не трогать теперь.
		break
	end
end


local function DefaultDisplayFn(arr) --На вход особая структура: { data, param }. data - ссылка на элемент MY_DATA, a param - ссылка на массив п.
	if arr.data == nil then
		return arr.param_str --Выводим строку в том виде, в каком пришла от сервера (без первого символа)
	end
	if arr.data.sign ~= nil and (tonumber(arr.param[1]) or -1) >= 0 then
		arr.param[1] = "+" .. tostring(arr.param[1])
	end
	local s = arr.data.desc .. " " .. tostring(arr.param[1])
	if arr.data.percent ~= nil then
		s = s .. "%"
	end
	return s
end
CallDefaultDisplayFn = DefaultDisplayFn --Для языкового модуля.

local function DefaultFraction(arr) --Типичный вывод: "Название: cur / max"
	local cur,mx = arr.param[1], arr.param[2]
	return arr.data.desc .. " " .. cur .. " / " .. mx 
end

MY_DATA.hp.fn = function(arr) --Формирование строки на клиенте.
	local cur,mx = arr.param[1], arr.param[2]
	return cur .. " / " .. mx 
end
MY_DATA.owner.fn = function(arr) --Формирование строки на клиенте
	return arr.data.desc .. " " .. arr.param_str  --Игнорируем запятые, если есть в имени.
end
MY_DATA.loyal.fn = function(arr) --Формирование строки на клиенте
	if (tonumber(arr.param[1]) or 0) > 9000 then
		return arr.data.desc .. " " .. SHOWME_STRINGS.loyal 
	end
	return DefaultDisplayFn(arr) --Игнорируем запятые, если есть в имени.
end

MY_DATA.water.fn = DefaultFraction
MY_DATA.water_poisoned.fn = function(arr)
	return arr.data.desc
end

MY_DATA.timer.fn = function(arr)
	local total = _G.tonumber(arr.param[1])
	local name = arr.param[2]
	local hours = math.floor(total * 0.0002777777777777) --целое кол-во часов. 1/3600
	local mins = math.floor((total - (hours * 3600)) * 0.01666666666666) --целое. 1/60
	local secs = total - (hours * 3600) - (mins * 60) --тоже должно быть целым.
	return arr.data.desc
		.. (name and (' "'..name..'": ') or ': ')
		.. (hours > 0 and (hours .. ':'
			.. (mins > 9 and mins or ('0' .. mins)) .. ':'
			.. (secs > 9 and secs or ('0' .. secs))
		) or ( mins .. ':'
			.. (secs > 9 and secs or ('0' .. secs))
		))
end
MY_DATA.durability.fn = DefaultFraction


MY_DATA.sanity_character.percent = true
MY_DATA.sanity.sign = true
MY_DATA.dmg_character.percent = true
MY_DATA.dmg_character.sign = true
MY_DATA.speed.percent = true
MY_DATA.speed.sign = true
MY_DATA.temperature.sign = true
MY_DATA.armor.percent = true
MY_DATA.armor_character.percent = true
MY_DATA.armor_character.sign = true
MY_DATA.waterproof.percent = true
MY_DATA.dmg_bonus.sign = true

MY_DATA.watergainspeed.sign = true
MY_DATA.sip.sign = true

MY_DATA.trade_gold.sign = true
MY_DATA.trade_rock.sign = true

local AOS_Temperature_fn --Клиентская функция конвертирования игровых юнтов температуры в визуальные.
local is_Fahrenheit = false
local function ConvertTemperature(val)
	if not val then
		return "???"
	end
	if is_Fahrenheit then
		return math.floor(1.8*(val) + 32.5).."\176F" 
	else
		return math.floor(val/2 + 0.5) .. "\176C"
	end
end

MY_DATA.temperature.sign = true
MY_DATA.temperature.fn = function(arr)
	arr.param[1] = AOS_Temperature_fn and AOS_Temperature_fn(tonumber(arr.param[1])) or ConvertTemperature(tonumber(arr.param[1]))
	return DefaultDisplayFn(arr)
end

--Возвращает полноценное название предмет. Берет из локальных переменных, так что должна быть переведено.
local function GetPrefabFancyName(prefab)
	local product = tostring(prefab or "nil")
	return _G.STRINGS.NAMES[string.upper(product)] or product
end

MY_DATA.cookpot.fn = function(arr)
	return arr.data.desc .. " " .. GetPrefabFancyName(arr.param[1])
end
MY_DATA.food.fn = function(arr)
	local hg,sn,hp = arr.param[1],arr.param[2],arr.param[3]
	if food_order <= 1 then
		if food_style <= 1 then
			return MY_DATA.hunger.desc .. " "..hg.." / "..MY_DATA.sanity.desc.." "..sn.." / "..MY_DATA.health.desc.." "..hp
		else
			return hg.." / "..sn.." / "..hp
		end
	else
		if food_style <= 1 then
			return MY_DATA.health.desc.." "..hp.." / "..MY_DATA.hunger.desc.." "..hg.." / "..MY_DATA.sanity.desc.." "..sn
		else
			return hp.." / "..hg.." / "..sn
		end
	end
end
MY_DATA.units_of.fn = function(arr)
	local s = FOOD_TAGS[arr.param[2]]
	s = s ~= "" and s or arr.param[2] --If translation exists.
	if arr.param[1] == "1" then
		return SHOWME_STRINGS.units_1 .. s
	else
		return arr.param[1] .. SHOWME_STRINGS.units_many .. s
	end
end
MY_DATA.uses_of.fn = function(arr)
	if arr.param[1] == "1" then
		return SHOWME_STRINGS.uses_1 .. arr.param[2]
	else
		return arr.param[1] .. SHOWME_STRINGS.uses_many .. arr.param[2]
	end
end
MY_DATA.perish.fn = function(arr)
	return arr.data.desc .. " " .. arr.param[1] .. SHOWME_STRINGS.days
end

local PerishFunction = function(arr)
	return MY_DATA.perish.fn(arr) --Redirect to perish function with another description - "Will die in..."
end

MY_DATA.will_die.fn = PerishFunction
MY_DATA.will_dry.fn = PerishFunction --Здесь тоже просто делаем ссылку.
MY_DATA.grow_in.fn = PerishFunction
MY_DATA.just_time.fn = PerishFunction

local CONST_COUNT = { ['1'] = '(1)', ['2'] = '(2)', ['3'] = '(3)', ['4'] = '(4)', ['5'] = '(5)', ['6'] = '(6)', ['7'] = '(7)',
	 ['8'] = '(8)', ['9'] = '(9)', ['10'] = '(10)', ['11'] = '(11)', ['12'] = '(12)', ['13'] = '(13)', ['14'] = '(14)', 
	 ['15'] = '(15)', ['16'] = '(16)', ['17'] = '(17)', ['18'] = '(18)', ['19'] = '(19)', ['20'] = '(20)', 
	 ['21'] = '(21)', ['22'] = '(22)', ['23'] = '(23)', ['24'] = '(24)', ['25'] = '(25)', 
	 ['26'] = '(26)', ['27'] = '(27)', ['28'] = '(28)', ['29'] = '(29)', ['30'] = '(30)', 
	 ['31'] = '(31)', ['32'] = '(32)', ['33'] = '(33)', ['34'] = '(34)', ['35'] = '(35)', 
	 ['36'] = '(36)', ['37'] = '(37)', ['38'] = '(38)', ['39'] = '(39)', ['40'] = '(40)', 
	 ['0'] = '', --Если нет конмонента stackable, то не показываем количество вообще.
}

--Особая функция вида "ПРОДУКТ: ВРЕМЯ"
--param[1] - продукт
--param[2] - оставшееся количество дней или nil
MY_DATA.perish_product.fn = function(arr)
	local name = GetPrefabFancyName(arr.param[1]) .. (
		CONST_COUNT[arr.param[2]] or ('(' .. tostring(arr.param[2]) .. ')')
	)
	if not arr.param[3] then --Время не указано.
		return name
	end
	arr.param[1] = arr.param[3] --dirty hack!!! Preparing for using in PerishFunction
	--arr.param[2] = nil --need not
	return name.. ":" .. PerishFunction(arr) -- desc == ""
end

MY_DATA.heal.sign = true
MY_DATA.fuel.percent = true
MY_DATA.obedience.percent = true
MY_DATA.domest.percent = true
--MY_DATA.fuel.percent = true
--MY_DATA.fuel.percent = true
--MY_DATA.fuel.percent = true

MY_DATA.crop.fn = function(arr) -- Product: percent
	return GetPrefabFancyName(arr.param[1]) .. ": " .. tostring(arr.param[2]) .. "%"
end

--MY_STRINGS_OVERRIDE = nil

--Сокрытие некоторой (назойливой) информации.
if show_food_units == 0 or show_food_units == 2 then
	MY_DATA.units_of.hidden = true --Work on client only.
end


function UpdateNewLanguage()
	--print(MY_STRINGS_OVERRIDE)
	if MY_STRINGS_OVERRIDE ~= nil then --Меняем локальный перевод (в т.ч. для хоста).
		for k,tr in pairs(MY_STRINGS_OVERRIDE) do
			local data = MY_DATA[k]
			if data ~= nil then
				data.desc = tr
			--else MY_STRINGS[k] = {tr}
			end
		end
	end
	--print(MY_STRINGS.aggro[1])
end


do --Пытаемся определить язык и загрузить соответствующий файл
	local support_languages = { ru = true, chs = true, cht = true, br = true, pl = true,
		tw="cht", zh_cn="chs", ch="chs", kr=true, ko="kr", es = true,} 
	--For override: name=file. Example: ,cht="chs",
	
	AddPrefabPostInit("world",function(inst)
		local lang = GetModConfigData("lang", true) or "auto"
		if lang == "auto" then
			lang = GetModConfigData("lang") or "auto"
		end
		if lang == "auto" then
			lang = _G.LanguageTranslator.defaultlang --print("LANG=",lang)
			if _G.STRINGS.ACTIONS.PLAY == "\253\253\253\253\253\253\253\253\253\253\6\196\232\253\253\253\253\253\253\253\253\1\205" then
				lang = "kr" --Korean
			end
		end
		if type(lang) ~= "string" or lang == "en" then
			return
		end
		lang = lang:lower()
		if support_languages[lang] ~= nil then
			if support_languages[lang] ~= true then --алиас
				lang = support_languages[lang]
			end
			modimport("showme_"..lang..".lua")
		end
		UpdateNewLanguage()
	end)
end



--Попытка определить клиентские моды и отправить их через RPC
AddModRPCHandler("ShowMe","AOS",function(inst)
	--Вызов этой функции означает отключение подсказок на деревьях для данного игрока.
	inst.has_AlwaysOnStatus = true
end)

--Учитывать ли статус порчи для конкретного пользователя.
AddModRPCHandler("ShowMe","Estimate",function(inst)
	--Вызов этой функции означает отключение подсказок на деревьях для данного игрока.
	inst.should_Estimate_Stale = true
end)

--local AOS_Temperature_fn --Функция конвертирования игровой температуры у клиента. (Объявлена ранее)

--Залезаем в клиентский интерфейс с целью уменьшить показываемую температуру
if CLIENT_SIDE then
	--Делаем самую правильную в мире функцию "AddPlayersPostInit" 
	--GetGlobal("mods",{})
	--if not _G.rawget(_G,"mods") then _G.rawset(_G,"mods",{}) end
	if not mods.player_preinit_fns then
		mods.player_preinit_fns={}
		--Dirty hack
		local old_MakePlayerCharacter = _G.require("prefabs/player_common")
		local function new_MakePlayerCharacter(...)
			local inst=old_MakePlayerCharacter(...)
			for _,v in ipairs(mods.player_preinit_fns) do
				v(inst)
			end
			return inst
		end
		_G.package.loaded["prefabs/player_common"] = new_MakePlayerCharacter
	end

	local function AddPlayersPreInit(fn)
		table.insert(mods.player_preinit_fns,fn)
	end

	local player_postinit_fns = {}
	local function AddPlayersPostInit(fn) -- <<<--------- Вот она!
		table.insert(player_postinit_fns,fn)
	end

	local done_players = {}
	AddPlayersPreInit(function(inst)
		local s = inst.prefab or inst.name
		if not done_players[s] then
			done_players[s] = true
			AddPrefabPostInit(s,function(inst)
				for _,v in ipairs(player_postinit_fns) do
					v(inst)
				end
			end)
		end
	end)

	local player_afterinit_fns = {}
	function AddPlayersAfterInit(fn) --Нулевой таймер после
		table.insert(player_afterinit_fns,fn)
	end
	AddPlayersPostInit(function(inst) --Задаем нулевой таймер
		if #player_afterinit_fns > 0 then
			inst:DoTaskInTime(0,function(inst)
				for i=1,#player_afterinit_fns do
					player_afterinit_fns[i](inst)
				end
			end)
		end
	end)
	
	--А вот и не угадали. Не трогаем интерфейс. Вместо этого определяем настройки и лишь в случае их отсутствия трогаем интерфейс.
	
	--Проверка температуры. Обратный формат и приведение к универсальному Цельсию.
	local tonumber = _G.tonumber
	local function FixTemperature(s)
		if type(s) ~= "string" then
			return s --Вообще не строка. Что за?
		end
		--На конце может быть один из конкретных вариантов строки.
		local sep = s:find("\176",1,true)
		if not sep then
			return s --Разделитель не найден.
		end
		local pre, pst = tonumber(s:sub(1,sep-1)), s:sub(sep+1)
		if not pre then
			return s --Первая часть строки не является числом.
		end
		if pst == "C" then
			return s --Это Цельсий. Не надо конвертировать.
		elseif pst == "F" then
			--Фаренгейт. Сложный случай. В целом тоже не надо трогать. Но меняем локальные настройки на Фаренгейт.
			is_Fahrenheit = true
			return s
		else
			--Иначе (по умолчанию) считаем, что температура в игровых единицах. Исправляем это дело на Цельсий.
			return ConvertTemperature(pre) --TODO: Есть небольшая погрешность в уменьшение. При температуре 5.1 покажется цифра 2, хотя должна 3.
		end
	end
	
	local AOS_UNITS, AOS_UNIT
	
	--Функция исправления интерфейса на клиенте. Запускается для всех игроков при появлении.
	local function FixClient(inst)
		if inst ~= _G.ThePlayer then
			return
		end
		
		--По-хакерски добавляем отправку настроек предпочтения показа еды
		if food_estimation == 1 then
			--print("SendRPC")
			SendModRPCToServer(MOD_RPC.ShowMe.Estimate)
		end
		
		local status = inst.HUD and inst.HUD.controls and inst.HUD.controls.status
		if not status then
			print("ERROR SHOW_ME: Can't fix client side status!")
			return
		end
		local AOS
		AOS_UNITS = TUNING.COMBINED_STATUS_UNITS
		AOS_UNIT = TUNING.COMBINED_STATUS_UNIT
		if (AOS_UNITS and AOS_UNIT and AOS_UNITS[AOS_UNIT]) then
			AOS = true
			AOS_Temperature_fn = AOS_UNITS[AOS_UNIT]
		else --older version of AOS
			if status.temperature then
				AOS = true
				local old_SetString = status.temperature.num.SetString
				status.temperature.num.SetString = function(self,s)
					return old_SetString(self,FixTemperature(s))
				end
			end
			if status.worldtemp then
				AOS = true
				local old_SetString = status.worldtemp.num.SetString
				status.worldtemp.num.SetString = function(self,s)
					return old_SetString(self,FixTemperature(s))
				end
			end
		end
		if AOS then
			SendModRPCToServer(MOD_RPC.ShowMe.AOS)
		end
	end
	AddPlayersAfterInit(FixClient) 
end

-------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------- HOST -----------------------------------------------------------------
--Works only on host.
if TheNet and TheNet:GetIsServer() then
require = _G.require





--Detecting enabled mods
local mod_names	 --названия всех модов (чтобы не дергать джвижок)
local mod_names_nover --названия модов с обрезанной версией (если смысловая часть достаточно длинная)
mod_names = {}
mod_names_nover = {}
local function GetAllModNames()
	if not (_G.KnownModIndex and _G.KnownModIndex.savedata and _G.KnownModIndex.savedata.known_mods
		and _G.ModManager and _G.ModManager.enabledmods)
	then
		return
	end
	local folders = {} --ассоциативный массив включенных модов (по папкам)
	for _,v in ipairs(_G.ModManager.enabledmods) do
		folders[v]=true
	end
	for folder, mod in pairs(_G.KnownModIndex.savedata.known_mods) do
		local name = mod.modinfo.name
		if name then
			mod_names[name]=folders[folder] and true or false
			--print("NEW_MOD: "..name.." "..tostring(mod_names[name]))
			local s=string.match(name,"^(.-)([0-9%._ ]+)$")
			if s then
				mod_names_nover[s]=name --обрезаем номер версии с конца
				--вместо true сохраняем полное имя мода (чтобы идентифицировать его в системе)
			else
				mod_names_nover[name]=name
			end
		end
	end
end
GetAllModNames()
mods.mod_names = mod_names
mods.mod_names_nover = mod_names_nover


local function SearchForModsByName()
	if mods.active_mods_by_name then
		return --Уже проинициализировано. Либо полная несовместимость.
	end
	mods.active_mods_by_name = {}
	if not (_G.KnownModIndex and _G.KnownModIndex.savedata and _G.KnownModIndex.savedata.known_mods) then
		print("ERROR COMMON LIB: Can't find KnownModIndex!")
		return
	end
	for name,mod in pairs(_G.KnownModIndex.savedata.known_mods) do
		if (mod.enabled or mod.temp_enabled or _G.KnownModIndex:IsModForceEnabled(name)) --Мод активен
			and not mod.temp_disabled --И не отключен
		then
			local real_name = mod.modinfo.name
			if real_name == nil then
				print("SHOW_ME ERROR: real_name of a mod is nil,",tostring(name))
				--TODO: error if TUNING.STAR_DEBUG 
			else
				mods.active_mods_by_name[real_name] = true
			end
		end
	end
end	
SearchForModsByName()
	
local is_HealthInfo = mods.active_mods_by_name["Health Info"] --Check it to decide, is there a reason to show hp in description.
local is_DisplayFoodValues = mods.active_mods_by_name["Display food values"]
--TODO: Эта проверка нужна на клиенте!
local is_AlwaysOnStatus = mods.active_mods_by_name["Combined Status"] or mods.active_mods_by_name["Always On Status"]
--_G.arr(mod_names)

local cooking = require("cooking")
local ing = cooking.ingredients
--local config = GetModConfigData("message_style") or 1

--local backdoors = GetGlobal("backdoors",{})
--backdoors.ShowMe={MY_STRINGS=MY_STRINGS} --welcome for translaters!
--for k,v in pairs(MY_STRINGS) do
--	MY_STRINGS[k]=v[config] --overwrites with string config values --Нельзя заменять, т.к. мы теперь используем ключи.
--end

--Добавляет строку в конец дескрипшена
--desc - исходная строка (копия), val - готовое значение, txt - ключ описания, no_table - если true, то ключ используется как строка.
--[[local function cn(desc,val,txt,no_table)
	if no_table then
		return (desc=="" and "" or (desc.."\n"))..txt.." "..tostring(val)
	elseif MY_STRINGS[txt] ~= nil and MY_STRINGS[txt][2] ~= nil then
		return (desc=="" and "" or (desc.."\n"))
			..MY_STRINGS[txt][2]
			..tostring(val)
	else
		return (desc=="" and "" or (desc.."\n"))..val.." (bug?)"
	end
end--]]

local ww
AddPrefabPostInit("world",function(inst)
	ww = inst.state
end)

local function GetPerishTime(inst,c)
	local modifier = 1
	local owner = c.inventoryitem and c.inventoryitem.owner or nil
	if owner == nil and c.occupier ~= nil then
		owner = c.occupier:GetOwner() --Для птичек?
	end

	if owner ~= nil then
		if owner:HasTag("fridge") then
			if inst:HasTag("frozen") and not owner:HasTag("nocool") and not owner:HasTag("lowcool") then
				modifier = TUNING.PERISH_COLD_FROZEN_MULT
			else
				modifier = TUNING.PERISH_FRIDGE_MULT
			end
		elseif owner:HasTag("spoiler") then
			modifier = TUNING.PERISH_GROUND_MULT 
		--elseif owner:HasTag("cage") and inst:HasTag("small_livestock") then
		--	modifier = TUNING.PERISH_CAGE_MULT
		end
	else
		modifier = TUNING.PERISH_GROUND_MULT 
	end

	if inst:GetIsWet() then
		modifier = modifier * TUNING.PERISH_WET_MULT
	end

	
	if ww.temperature < 0 then
		if inst:HasTag("frozen") and not c.perishable.frozenfiremult then
			modifier = TUNING.PERISH_COLD_FROZEN_MULT
		else
			modifier = modifier * TUNING.PERISH_WINTER_MULT
		end
	end

	if c.perishable.frozenfiremult then
		modifier = modifier * TUNING.PERISH_FROZEN_FIRE_MULT
	end

	if ww.temperature > TUNING.OVERHEAT_TEMP then
		modifier = modifier * TUNING.PERISH_SUMMER_MULT
	end

	if c.perishable.localPerishMultiplyer then
		modifier = modifier * c.perishable.localPerishMultiplyer
	end

	modifier = modifier * TUNING.PERISH_GLOBAL_MULT
	
	local old_val = c.perishable.perishremainingtime
	if old_val ~= nil then
		local delta = old_val / modifier
		return delta
	end
end



--Новая функция
local desc_table
local function cn(key,param1,param2,param3)
	local data = MY_DATA[key]
	if data == nil then
		return
	end
	if param1 == nil then
		table.insert(desc_table, data.sym)
		return 
	end
	if param2 == nil then
		table.insert(desc_table, data.sym ..tostring(param1))
		return 
	end
	if param3 == nil then
		table.insert(desc_table, data.sym ..tostring(param1) .. "," ..tostring(param2))
		return
	end
	table.insert(desc_table, data.sym ..tostring(param1) .. "," ..tostring(param2) .. "," ..tostring(param3))
	return 
end

--Hack for LRP
--[[if _G.rawget(_G,"RussificationVersion") and _G.RussificationVersion<"3.6" then
	local player = _G.STRINGS.CHARACTERS.GENERIC.DESCRIBE.PLAYER
	if player then --Add some spaces to strings. So they are incorruptible.
		player.GENERIC = "It  is %s!"
		player.ATTACKER = "That %s looks  shifty..."
		player.MURDERER = "Murderer !"
		player.REVIVER = "%s, friend  of ghosts."
		player.GHOST = "%s could use a	heart."
	end
end --]]


--nice round function
local round2=function(num, idp)
	return _G.tonumber(string.format("%." .. (idp or 0) .. "f", num))
end

local is_admin
local last_user_talbe = {}
--Проверяет, является ли чел админом.
local function IsAdmin(viewer)
	if is_admin ~= nil then
		return is_admin
	end
	if not (viewer and viewer.userid) then
		return false
	end
	for i=1,#last_user_talbe do
		local user = last_user_talbe[i]
		if user.userid == viewer.userid then
			is_admin = user.admin or false
			return is_admin
		end
	end
	last_user_talbe = _G.TheNet:GetClientTable()
	for i=1,#last_user_talbe do
		local user = last_user_talbe[i]
		if user.userid == viewer.userid then
			is_admin = user.admin or false
			return is_admin
		end
	end
end


local function name_by_id(userid)
	for i,v in ipairs(_G.AllPlayers) do
		if v.userid == userid then
			return v.name
		end
	end
	return "---Unknown---"
end
--GetGlobal("name_by_id",name_by_id)

--local TemperatureFormatLocal

local SPICIAL_STRUCTURES = {
	campfire = true, coldfire = true,
}

--Основная функция получения описания.
function GetTestString(item,viewer) --Отныне форкуемся от Tell Me, ибо всё сложно.
	--line_cnt = 0
	desc_table = {} --старый desc отменяется
			
	is_admin = nil
	local prefab = item.prefab
	local c=item.components
	local has_owner = false --Выводим инфу о владельце лишь ОДИН раз!
	if (prefab=="evergreen" or prefab=="deciduoustree") and not viewer.has_AlwaysOnStatus then
		--if not is_AlwaysOnStatus then --TODO: Do not check! NB!
			local w=_G.TheWorld.state
			local tt=round2(w.temperature,1)
			if w.iswinter then cn("S1")
			elseif w.issummer then cn("S2")
			elseif w.isspring then cn("S3")
			elseif w.isautumn then cn("S4")
			end
			cn("remaining_days",w.remainingdaysinseason)
			cn("temperature",tt)
				--..MY_STRINGS["remaining_days"][2]..": "..w.remainingdaysinseason.."\n"
				--.."t "..(tt>=0 and "+" or "")..tt
		--end
	elseif c.health and not item.grow_stage then --Health, Hunger, Sanity Bar
		local h=c.health
		--cheat
		if item.is_admin then
			cn("is_admin")
			return desc_table[1]
		end
		
		if not is_HealthInfo then --c.health
			local mx=math.ceil(h.maxhealth-h.minhealth)
			local cur=math.ceil(h.currenthealth-h.minhealth)
			if cur>mx then cur=mx end
			cn("hp",cur,mx)
		end

		if c.hunger then
			local val = c.hunger:GetPercent()
			--Либо голода мало, либо это вообще не игрок.
			if (c.grogginess and val <= 0.5) or (not c.grogginess and (val > 0 or prefab ~= "beefalo")) then
				cn("hunger",round2(c.hunger.current,0))
			end
		elseif c.perishable ~= nil and c.perishable.updatetask ~= nil then --Here "Perishable" means "Hunger".
			local time = GetPerishTime(item, c)
			if time ~= nil then
				cn("will_die",round2(time/TUNING.TOTAL_DAY_TIME,1))
			end
		end
		if c.sanity and c.sanity:GetPercent()<=0.5 then
			local sanity = round2(math.floor(c.sanity:GetPercent()*100+0.5),1)
			cn("sanity_character",sanity)
		end
		if c.follower then
			if c.follower.leader and c.follower.leader:IsValid() and c.follower.leader:HasTag("player")
				and c.follower.leader.name and c.follower.leader.name ~= ""
			then
				cn("owner",c.follower.leader.name)
				has_owner = true
			end
			if c.follower.maxfollowtime then
				mx = c.follower.maxfollowtime
				cur = math.floor(c.follower:GetLoyaltyPercent()*mx+0.5)
				if cur>0 then
					cn("loyal",cur,mx)
				end
			end
		end
		--[[if c.locomotor and type(c.locomotor.walkspeed)=="number" then
			local speed = (c.locomotor.walkspeed + (type(c.locomotor.bonusspeed)=="number" and c.locomotor.bonusspeed or 0))
				/ TUNING.WILSON_WALK_SPEED
			if speed>1.01 or speed<0.99 then
				desc = cn(desc,"x"..round2(speed,2),"speed")
			end
		end --]]
		if item.kills and item.kills>0 then
			cn(item.kills==1 and "kill" or "kills",item.kills)
		end
		if item.aggro and item.aggro>0 then
			cn("aggro",item.aggro)
		end
		--Процент увеличения урона. Например, у Вигфрид +25%
		if c.combat and c.combat.damagemultiplier and c.combat.damagemultiplier ~= 1 then
			local perc = c.combat.damagemultiplier - 1
			cn("dmg_character",round2(perc*100,0))
		end
		--Урон
		if c.combat and c.combat.defaultdamage and c.combat.defaultdamage > 0 then
			--Игнорируем всех, чья сила равна 10 или меньше.
			local com = c.combat
			local dmg = com.defaultdamage * (com.playerdamagepercent or 1)
			cn("strength", math.floor( dmg + 0.5))
			if com.areahitdamagepercent then --AoE
				cn("aoe", math.floor( dmg * com.areahitdamagepercent + 0.5))
			end
		end
		if h.absorb~=0 or h.playerabsorb~=0 then
			local perc = 1-(1-h.absorb)*(1-h.playerabsorb)
			cn("armor_character",round2(perc*100,0))
		end
		if item.asunaheal_score and item.prefab == "asuna" and TUNING.ASUNA_HEAL_SCORE_SWORD
			and item.asunaheal_score < TUNING.ASUNA_HEAL_SCORE_SWORD
		then
			local asuna_proof = round2(math.floor((item.asunaheal_score/TUNING.ASUNA_HEAL_SCORE_SWORD)*100+0.5),0)
			if asuna_proof > 99 then
				asuna_proof = 99
			end
			table.insert(desc_table, "$Asuna Proof: "..asuna_proof.."%")
		end
		--inst.components.domesticatable:GetObedience()
		if c.domesticatable ~= nil then
			if c.domesticatable.GetObedience ~= nil then
				local obedience = c.domesticatable:GetObedience()
				if obedience ~= 0 then
					cn("obedience",round2(obedience*100,0))
				end
			end
			if c.domesticatable.GetDomestication ~= nil then
				local domest = c.domesticatable:GetDomestication()
				if domest ~= 0 then
					cn("domest",round2(domest*100,0))
				end
			end
		end
	else --elseif prefab~="rocks" and prefab~="flint" then --No rocks and flint
		--Part 1: primary info
		if c.stewer and c.stewer.product and c.stewer.IsCooking and c.stewer:IsCooking() then
			local tm=round2(c.stewer.targettime-_G.GetTime(),0)
			if tm<0 then tm=0 end
			cn("cookpot", c.stewer.product)
			cn("sec",tm)
		end
		if c.cooldown and c.cooldown.GetTimeToCharged and not c.cooldown.charged then
			local timer = round2(c.cooldown:GetTimeToCharged(),0)
			cn("cooldown", timer)
		end
		--Part 2: secondary info
		if c.armor and c.armor.absorb_percent and type(c.armor.absorb_percent)=="number" then
			local r=c.armor.absorb_percent
			cn("armor",round2(r*100,0))
			--Support of absorption mod.
			if item.phys and (item.phys.blunt or item.phys.pierc or item.phys.slash) then
				local p = item.phys
				cn("absorb",(p.blunt or 0).." / "..(p.pierc or 0).." / "..(p.slash or 0))
			end
			if c.armor.condition and c.armor.condition > 0 and c.armor.maxcondition then
				cn("durability", math.floor(c.armor.condition + 0.5), math.floor(c.armor.maxcondition + 0.5))
			end
		end
		if item.damage and type(item.damage)=="number" and item.damage>0 then
			cn("dmg",round2(item.damage,1))
		elseif c.weapon ~= nil and c.weapon.damage and type(c.weapon.damage)=="number" and c.weapon.damage>0 then
			cn("dmg",round2(c.weapon.damage,1))
			--Support of absobtion mod.
			if item.phys_dmg then
				local p = item.phys_dmg == "blunt" and "Blunt" or (
					item.phys_dmg == "pierc" and "Piercing" or (
						item.phys_dmg == "slash" and "Slashing" or nil
					)
				)
				if p ~= nil then
					table.insert(desc_table, "$Type: "..p)
				end
			end
		elseif c.zupalexsrangedweapons ~= nil
			and c.zupalexsrangedweapons.GetArrowBaseDamage ~= nil
			and type(c.zupalexsrangedweapons.GetArrowBaseDamage) == "function"
		then
			local dmg = c.zupalexsrangedweapons:GetArrowBaseDamage()
			if dmg ~= nil and type(dmg) == "number" and dmg > 0 then
				cn("dmg",round2(dmg,1))
			end
		end
		if c.weapon and c.weapon.damage and type(c.weapon.attackrange)=="number" and c.weapon.attackrange>0.3 then
			cn("range",round2(c.weapon.attackrange,1))
		elseif c.projectile and c.projectile.damage and type(c.projectile.range)=="number" and c.projectile.range>0.3 then
			cn("range",round2(c.projectile.range,1))
		elseif c.combat and c.combat.damage and type(c.combat.attackrange)=="number" and c.combat.attackrange>2.5 then
			cn("range",round2(c.combat.attackrange,1))
		end
		if c.insulator and c.insulator.insulation and type(c.insulator.insulation)=="number" and c.insulator.insulation~=0 then
			if c.insulator.SetInsulationEx then --ServerMod
				local winter,summer = c.insulator:GetInsulationEx()
				if winter~=0 then
					cn("warm",round2(winter,0))
				end
				if summer~=0 then
					cn("summer",round2(summer,0))
				end
			elseif c.insulator.GetInsulation then
				local insul,typ = c.insulator:GetInsulation()
				if typ == _G.SEASONS.WINTER then
					cn("warm",round2(insul,0))
				elseif typ == _G.SEASONS.SUMMER then
					cn("summer",round2(insul,0))
				end
			end
		end
		if c.dapperness and c.dapperness.dapperness and type(c.dapperness.dapperness)=="number" and c.dapperness.dapperness~=0 then
			local sanity = round2(c.dapperness.dapperness*54,1)
			cn("sanity",sanity)
		elseif c.equippable and c.equippable.dapperness and type(c.equippable.dapperness)=="number" and c.equippable.dapperness~=0 then
			local sanity = round2(c.equippable.dapperness*54,1)
			cn("sanity",sanity)
		end
		if c.equippable and c.equippable.walkspeedmult and c.equippable.walkspeedmult ~= 1 then
			local added_speed = math.floor((c.equippable.walkspeedmult - 1)*100+0.5)
			cn("speed",added_speed)
		end
		if c.dapperness and c.dapperness.mitigates_rain and item.prefab ~= "umbrella" then
			cn("waterproof","90")
		elseif item.protect_from_rain then
			cn("waterproof",round2((item.protect_from_rain)*100,0))
		elseif c.waterproofer then
			local effectiveness = _G.tonumber(c.waterproofer.effectiveness) or 0
			if effectiveness ~= 0 then
				cn("waterproof",round2((effectiveness)*100,0))
			else
				--desc = (desc=="" and "" or (desc.."\n")).."Waterproofer"
			end
		end
		--if c.striker and c.striker.chance and type(c.striker.chance) == "number" then
		--	desc = cn(desc,round2((c.striker.chance)*100,0).."%","striker")
		--end
		--if c.tinder and c.tinder.tinder and c.tinder.GetTinder then
		--	local power = c.tinder:GetTinder()
		--	if power >= 0.005 then
		--		desc = cn(desc,round2(power*100,0).."%","tinder")
		--	end
		--end
		if c.edible and not is_DisplayFoodValues then
			local can_eat = false
			if viewer and viewer.components.eater then
				can_eat = viewer.components.eater:CanEat(item)
			end
			if can_eat then
				local should_Estimate_Stale = viewer and viewer.should_Estimate_Stale --client priority
				if not should_Estimate_Stale then
					should_Estimate_Stale = food_estimation ~= 0
				end
				local hp,hg,sn
				if should_Estimate_Stale and c.edible.GetSanity then
					--print("Estimate")
					hp=round2(c.edible:GetHealth(viewer),1)
					hg=round2(c.edible:GetHunger(viewer),1)
					sn=round2(c.edible:GetSanity(viewer),1)
				else
					--print("Not Estimate")
					hp=round2(c.edible.healthvalue,1)
					hg=round2(c.edible.hungervalue,1)
					sn=round2(c.edible.sanityvalue,1)
				end
				if viewer ~= nil and viewer.FoodValuesChanger ~= nil then --Особая функция, призвание которой - менять еду при съедании.
					--print("+")
					--Правда, здесь мы можешь слегка подсмотреть ее результаты до поедания.
					local hp2, hg2, sn2 = viewer:FoodValuesChanger(item)
					if sn2 ~= nil then
						--print("++")
						hp=round2(hp2,1)
						hg=round2(hg2,1)
						sn=round2(sn2,1)
					end
				end
				if hp > 0 then
					hp = "+" .. tostring(hp)
				end
				if hg > 0 then
					hg = "+" .. tostring(hg)
				end
				if sn > 0 then
					sn = "+" .. tostring(sn)
				end
				cn("food",hg,sn,hp)
			end
		end
		if c.perishable ~= nil and c.perishable.updatetask ~= nil then
			local time = GetPerishTime(item, c)
			if time ~= nil then
				cn("perish",round2(time/TUNING.TOTAL_DAY_TIME,1))
			end
		end
		if ing[prefab] and show_food_units ~= 2 then -- ==2 means that food info is forbidden on the server.
			for k,v in pairs(ing[prefab].tags) do
				if k~="precook" and k~="dried" then
					cn("units_of",v,k)
				end
			end
		end
		if c.healer then
			cn("heal",round2(c.healer.health,1))
		end
		--[[if item.grow_stage and type(item.grow_stage) == "number" then --Support Clan System mod
			local val = math.floor(item.grow_stage+0.5)
			desc = cn(desc,tostring(val).."%","power")
			if IsAdmin(viewer) then
				desc = cn(desc,item.show_stage,"show_stage",true)
				desc = cn(desc,item.grow_stage,"grow_stage",true)
				desc = cn(desc,item.active,"active",true)
				desc = cn(desc,item.fuel,"fuel",true)
			end
		else--]]
		if c.finiteuses then
			local mult = 1
			if c.finiteuses.consumption then
				for k,v in pairs(c.finiteuses.consumption) do
					mult = 1/v
				end
			end
			local cur = math.floor(c.finiteuses.current * mult + 0.5)
			if c.finiteuses.current > cur then
				cur = cur + 1
			end
			cn("uses_of",cur,math.floor(c.finiteuses.total * mult + 0.5))
			--desc = (desc=="" and "" or (desc.."\n"))..cur.." use"..(cur~=1 and "s" or "").." of "..c.finiteuses.total
		end
		if c.temperature and c.temperature.current and type(c.temperature.current) == "number" then
			cn("temperature",round2(c.temperature.current,1))
		end
		if c.fueled and c.fueled:GetPercent()>0 and (SPICIAL_STRUCTURES[prefab] or item:HasTag("structure")) then
			cn("fuel",round2(c.fueled:GetPercent()*100,0))
		end
		if c.instrument and type(c.instrument.range)=="number" and c.instrument.range>0.4 then
			cn("range",round2(c.instrument.range,0))
		end
		if c.crystallizable and c.crystallizable.formation --support of Krizor's mod
			and c.crystallizable.formation.thickness
			and type(c.crystallizable.formation.thickness)=="table"
			and c.crystallizable.formation.thickness.current
			and c.crystallizable.formation.thickness.current>0
		then
			cn("thickness",round2(c.crystallizable.formation.thickness.current,1))
		end
		if c.mine then
			if c.mine.nick then
				cn("owner",c.mine.nick)
				has_owner = true
			end
			--[[if c.mine.pret and viewer and viewer.userid then
				for k,v in pairs(c.mine.pret) do
					if k==viewer.userid then
						--desc = (desc=="" and "" or (desc.."\n")).."I can see it!"
						desc = cn(desc,v,"known",true)
						break
					end
				end
			end--]]
		end
		if not has_owner then
			if item.stealable and item.stealable.owner and item.stealable.owner ~= "_?\1" then
				cn("owner",item.stealable.owner)
				has_owner = true
			elseif item.owner and type(item.owner)=="string" and string.sub(item.owner,1,3) ~= "KU_" then
				--Мы не знаем, что за имя. Но это "владелец". Так что надо вывести. И это точно не user_id.
				cn("owner",item.owner)
				has_owner = true
			end
		end
		if c.occupiable then
			local item = c.occupiable:GetOccupant()
			if item then
				local c = item.components
				if c.perishable ~= nil and c.perishable.updatetask ~= nil then --Here "Perishable" means "Hunger".
					local time = GetPerishTime(item, c)
					if time ~= nil then
						cn("will_die",round2(time/TUNING.TOTAL_DAY_TIME,1))
					end
				end
			end
		end
		if c.dryer and c.dryer.IsDrying then
			if c.dryer:IsDrying() and c.dryer.GetTimeToDry then
				cn("will_dry",round2(c.dryer:GetTimeToDry()/TUNING.TOTAL_DAY_TIME,1))
				--if c.dryer:IsPaused() then
				--end
			elseif c.dryer.IsDone and c.dryer:IsDone() and c.dryer.GetTimeToSpoil then
				cn("perish",round2(c.dryer:GetTimeToSpoil()/TUNING.TOTAL_DAY_TIME,1))
			end
		end
		if c.saddler then --Седло и его параметры.
			if c.saddler.speedmult and c.saddler.speedmult ~= 0 then
				local added_speed = math.floor((c.saddler.speedmult - 1)*100 + 0.5) -- (1.4 - 1) == 0.4
				cn("speed",added_speed)
			end
			if c.saddler.bonusdamage and c.saddler.bonusdamage ~= 0 then
				cn("dmg_bonus",round2(c.saddler.bonusdamage,1))
			end
		end
		if c.timer and c.timer.timers then
			local get_time = GetTime()
			for name, data in pairs(c.timer.timers) do
				cn('timer', round2(data.end_time - get_time,0), name)
			end
		end
		if c.tradable then
			if c.tradable.goldvalue and c.tradable.goldvalue > 0 then
				cn("trade_gold", c.tradable.goldvalue)
			end
			if c.tradable.rocktribute and c.tradable.rocktribute > 0 then
				cn("trade_rock", c.tradable.rocktribute)
			end
		end
		------------------Check prefabs?----------------------
		if prefab=="panflute" then
			--desc = cn("power","10")
		elseif prefab=="blowdart_sleep" then
			--desc = cn(desc,"1","power")
		--[[elseif prefab=="pond" and item.targettime then
			local tm = item.targettime - _G.GetTime()
			if tm>0 then
				desc = "Broken "..cn(desc,tm,"sec")
			elseif item.broken then
				desc = (desc=="" and "" or (desc.."\n")).."Broken"
			end--]]
		elseif prefab=="pond" or prefab=="pond_mos" then
			cn(c.fishable.fishleft==1 and "fish" or "fishes",c.fishable.fishleft)
		elseif prefab=="aqvarium" and item.data then
			if item.data.seeds and item.data.seeds>0 then
				table.insert(desc_table, "$Seeds: "..tostring(item.data.seeds))
			end
			if item.data.meat and item.data.meat>0 then
				table.insert(desc_table, "$Meat: "..tostring(item.data.meat))
				--desc = cn(desc,item.data.meat,"Meat:",true)
			end
			local need_wet= item.data.need_wet or 60
			if item.data.wet and item.data.wet>0 and item.data.wet<need_wet then
				table.insert(desc_table, "$Water: "..tostring(round2(100*item.data.wet/need_wet).."%"))
				--desc = cn(desc,round2(100*item.data.wet/need_wet).."%","Water:",true)
			end
			if item.total_heat then
				local temp = item.total_heat/10 --+ _G.TheWorld.state.temperature
				if temp>40 then temp = 40 end
				if temp>=0 then
					cn("temperature",tostring(round2(temp,1)))
				end
			end
		end
		--Charges: lightning rod / lamp
		if item.chargeleft and item.chargeleft > 0 then	
			table.insert(desc_table, "$Days left: "..tostring(math.floor(item.chargeleft+0.5)))
		end
		--Mod support:
		if item.GetShowItemInfo then
			local custom1, custom2, custom3 = item:GetShowItemInfo()
			if custom1 then table.insert(desc_table, "$"..tostring(custom1)) end
			if custom2 then table.insert(desc_table, "$"..tostring(custom2)) end
			if custom3 then table.insert(desc_table, "$"..tostring(custom3)) end
		end
		if c.pickable and c.pickable.task then --Трава и ветки.
			local targettime = c.pickable.targettime
			if targettime then
				local delta = targettime - GetTime()
				if delta > 0 then
					cn("grow_in",round2(delta/TUNING.TOTAL_DAY_TIME,1)) --days
				end
			end
		end
		--[[if c.witherable then
			local time = GetTime()
			table.insert(desc_table, "$witherable: "
				..tostring(c.delay_to_time and (time-c.delay_to_time)) .. ", "
				..tostring(c.task_to_time and (time-c.task_to_time)) .. ", "
				..tostring(c.protect_to_time and (time-c.protect_to_time)) .. ", "
				..tostring(c.is_watching_rain)
			)
		end
		if c.diseaseable then
			local time = GetTime()
			table.insert(desc_table, "$diseaseable: "
				--..tostring(c._spreadtask and (time-c.delay_to_time)) .. ", "
				--..tostring(c.task_to_time and (time-c.task_to_time)) .. ", "
				--..tostring(c.protect_to_time and (time-c.protect_to_time)) .. ", "
				--..tostring(c.is_watching_rain)
			)
		end--]]
		--Грядки - это на самом деле высаженные всходы (мышка наводится на них поверх грядок).
		if c.crop and c.crop.product_prefab and c.crop.product_prefab and c.crop.growthpercent
			and type(c.crop.growthpercent) == 'number' and c.crop.growthpercent < 1
		then
			--Передаем названием продукта и процент созревания (до целых).
			cn("crop",c.crop.product_prefab,round2(c.crop.growthpercent*100,0))
		end
		--c.unwrappable.itemdata[1].prefab
		--c.unwrappable.itemdata[1].data.perishable.time - оставшееся время порчи в секундах.
		--	   .stackable.stack - количество
		if c.unwrappable and c.unwrappable.itemdata and type(c.unwrappable.itemdata) == 'table' then
			--По одной строке на каждый предмет.
			for i,v in ipairs(c.unwrappable.itemdata) do
				if v.prefab then
					--Пересылаем название префаба и количество дней.
					local delta = v.data and v.data.perishable and v.data.perishable.time
					local count = v.data and v.data.stackable and v.data.stackable.stack
					cn('perish_product', v.prefab, count or 0, delta and round2(delta/TUNING.TOTAL_DAY_TIME,1))
				end
			end
		end
		--Боченок рассола из мода "Pickle It!"
		if c.pickler and c.pickler.targettime then
			local delta = c.pickler.targettime - GetTime()
			cn('just_time', round2(delta/TUNING.TOTAL_DAY_TIME,1))
		end
		--Для мода "Thirst" проверяем компонент "cwater"
		if c.cwater then
			local w = c.cwater
			--Особо не заморачиваемся. Просто выводим то, что в нём есть.
			if w.current and w.max then
				cn('water',round2(w.current,0),round2(w.max,0))
			end
			if w.waterperdrink and type(w.waterperdrink)=="number" and w.waterperdrink ~= 0 then
				cn("sip",round2(w.waterperdrink,0))
			end
			if w.watergainspeed and type(w.watergainspeed)=="number" and w.watergainspeed ~= 0 then
				cn("watergainspeed",round2(w.watergainspeed,0))
			end
			if w.poisoned then
				cn("water_poisoned")
			end
		end
	end
	--Depending from weapon info:
	if viewer and type(viewer)=="table" and viewer.components and viewer.components.inventory then
		local weapon = viewer.components.inventory:GetEquippedItem(_G.EQUIPSLOTS.HANDS)
		if weapon then
			if weapon.prefab=="icestaff" and c.freezable then
				cn("resist",c.freezable.resistance)
			elseif (weapon.prefab=="blowdart_sleep" or weapon.prefab=="panflute") and c.sleeper then
				cn("resist",c.sleeper.resistance)
			end
		end
	end
	if item.inlove and item.inlove>0 then
		if prefab=="chester" then
			cn("love",item.inlove/10)
		else
			cn("love",item.inlove)
		end
	end
	--Additional
	--[[
	if prefab=="chester" then
		local name = name_by_id(self.inst.userid)
		desc = (desc=="" and "" or (desc.."\n")).."Owner: "..name
			.."\nuserid="..tostring(self.inst.userid)
			.."\nLeader: "..tostring(c.follower.leader)
		has_owner = true
	end
	if prefab=="chester_eyebone" then
		local name = name_by_id(self.inst.userid)
		desc = (desc=="" and "" or (desc.."\n")).."Owner: "..name
			.."\nuserid="..tostring(self.inst.userid)
		has_owner = true
	end
	--]]
	--print("GetTestString: "..tostring(item)..", "..tostring(viewer)..", "..tostring(desc))
	--for i=1,line_cnt do
	--	desc = desc .. "\n" --Поднимаем описание предмета, чтобы оно было НАД предметом. Но лучше это сделать на клиенте.
	--end
	
	return table.concat(desc_table,"\2")
end

--Main description function
--[[
AddComponentPostInit("inspectable", function(inst)
	local oldGetDesc = inst.GetDescription
	function inst:GetDescription(viewer)
		local desc = oldGetDesc(self,viewer)
		if self.inst and self.inst.components then
			local item = self.inst
			if type(desc)~="string" then
				desc=""
			end
			local desc = GetTestString(item,viewer,desc)
		end
		return desc
	end
end)
--]]

--[[
AddPrefabPostInit("wes",function(wes)
	if wes.components and wes.components.talker then
		local oldSay = wes.components.talker.Say
		function wes.components.talker:Say(script, time, noanim)
			local test=type(script)=="string" and string.match(script,"([0-9]+ / [0-9]+)$") or ""
			if test then
				script=test
			else
				script=""
			end
			oldSay(self, script, time, noanim)
		end
		wes.components.talker.special_speech = false
	end
end)

GetGlobal("fol",function()
	
end)
--]]


end


-------------------------------------------------------------------------------------------------------------
------------------------------------------- HOST & CLIENT AGAIN ---------------------------------------------





--Добавляем подсказку для игрока, через которую будем пересылать данные (всплывающий текст с инфой под именем предмета)
do
	--Функция возвращает подсказку, если она в точности совпадает с присланной информацией (в player_classified).
	--И возвращает подсказку, либо "".
	local function CheckUserHint(inst)
		local c = _G.ThePlayer and _G.ThePlayer.player_classified
		if c == nil then --Нет локального игрока или classified
			return ""
		end
		--c.showme_hint
		local i = string.find(c.showme_hint,';',1,true)
		if i == nil then --Строка имеет неправильный формат.
			return ""
		end
		local guid = _G.tonumber(c.showme_hint:sub(1,i-1))
		if guid ~= inst.GUID then --guid не совпадает (либо вообще nil)
			return ""
		end
		return c.showme_hint:sub(i+1)
	end
	if CLIENT_SIDE then
		--patching Get Display Name. Нужно только клиенту.
		--[[local old_GetDisplayName = _G.EntityScript.GetDisplayName
		_G.EntityScript.GetDisplayName = function(self)
			local old_name = old_GetDisplayName(self)
			if type(old_name) ~= "string" then
				return old_name
			end
			local str2 = CheckUserHint(self)
			return old_name .. str2
		end--]]
		
		--Разбираем случаи, когда нужно отправить guid об объекте под мышью.
		local old_inst --Запоминаем, чтобы не спамить один и тот же inst по несколько раз.
		--[[AddWorldPostInit(function(w)
			w:DoPeriodicTask(0.1,function(w)
				if _G.ThePlayer == nil then
					return
				end
				local inst = _G.TheInput:GetWorldEntityUnderMouse()
				if inst ~= nil then
					if inst == old_inst then
						return
					end
					old_inst = inst
					--Посылаем желаемую подсказку.
					SendModRPCToServer(MOD_RPC.ShowMeHint.Hint, inst.GUID, inst)
				end
			end)
		end)--]]
		
		local function UnpackData(str,div)
			local pos,arr = 0,{}
			-- for each divider found
			for st,sp in function() return string.find(str,div,pos,true) end do
				table.insert(arr,string.sub(str,pos,st-1)) -- Attach chars left of current divider
				pos = sp + 1 -- Jump past current divider
			end
			table.insert(arr,string.sub(str,pos)) -- Attach chars right of last divider
			return arr
		end		

		local save_target
		local last_check_time = 0 --последнее время проверки. Будет устаревать каждые 2 сек.
		local LOCAL_STRING_CACHE = {} --База данных строк, чтобы не обсчитывать замены каждый раз (правда, будет потихоньку пожирать память)
		AddClassPostConstruct("widgets/hoverer",function(hoverer) --hoverer=self
			local old_SetString = hoverer.text.SetString
			hoverer.text.SetString = function(text,str) --text=self
				--print(tostring(str))
				local target = _G.TheInput:GetHUDEntityUnderMouse()
				if target ~= nil then
					--target.widget.parent - это ItemTile
					target = target.widget ~= nil and target.widget.parent ~= nil and target.widget.parent.item --реальный итем (на клиенте)
				else
					target = _G.TheInput:GetWorldEntityUnderMouse()
				end
				--local lmb = hoverer.owner.components.playercontroller:GetLeftMouseAction()
				if target ~= nil then
					--print(tostring(target))
					--Проверяем совпадение с данными.
					local str2 = CheckUserHint(target)
					if str2 ~= "" then
						--Так, сначала чистим старую строку от переходов на новую строку. Мало ли какие там моды чего добавили.
						local cnt_newlines, _ = 0 --Считаем переходы строк в конце строки (совместимость с DFV)
						while cnt_newlines < #str do
							local ch = str:sub(#str-cnt_newlines,#str-cnt_newlines)
							if ch ~= "\n" and ch ~= " " then
								break
							end
							cnt_newlines = cnt_newlines + 1
						end
						--Очищаем строку от этого мусора
						if cnt_newlines > 0 then
							str = str:sub(1,#str-cnt_newlines)
						end
						--print(#str,"clear")
						--Очищаем строку от промежуточного мусора
						if string.find(str,"\n\n",1,true) ~= nil then
							str = str:gsub("[\n]+","\n")
						end
						--[[ --Теперь это не нужно, т.к. мы тупо добавляем 1 строку.
						if string.find(str,"\n",1,true) ~= nil then
							_,cnt_newlines = str:gsub("\n","\n") --Подсчитываем количество переходов внутри (если есть).
						else
							cnt_newlines = 0
						end
						--]]

						--Извлекаем данные из полученной упакованной строки.
						str2 = UnpackData(str2,"\2")
						local arr2 = {} --Формируем массив данных в удобоваримом виде.
						for i,v in ipairs(str2) do
							if v ~= "" then
								local param_str = v:sub(2)
								local data = { param = UnpackData(param_str,","), param_str=param_str }
								local my_s = MY_STRINGS[string.byte(v:sub(1,1))-64]
								if my_s ~= nil then
									data.data = MY_DATA[my_s.key]
								end
								table.insert(arr2,data)
							end
						end
						arr2.str2= str2
						--_G.rawset(_G,"arr2",arr2) --Для теста.
						--Формируем строку
						str2=""
						for i,v in ipairs(arr2) do
							if v.data ~= nil and v.data.fn ~= nil then
								str2=str2.. v.data.fn(v) .. "\n"
							elseif v.hidden == nil then
								str2=str2 .. DefaultDisplayFn(v) .. "\n"
							end
						end
						
						--_G.arr({inst=text.inst,hover=text.parent},5)
						--print("-----"..str.."-----")
						--local sss=""
						--for i=#str,#str-10,-1 do
						--	sss=sss..string.byte(str:sub(i,i))..", "
						--end
						--print("Chars: "..sss)
						--[[print(#str,"cut str")
						--В конце тоже убираем переход, если есть.
						if str:sub(#str,#str) == "\n" then
							str = str:sub(1,#str-1)
						end--]]
						--print(#str,"test cache")
						--print("count new cache")
						--print("newlines",#str2)
						
						str = str .. "\n" .. str2
						
						--Заново считаем количество оставшихся переходов строк
						--cnt_newlines = cnt_newlines + #arr2 + 1 --Всего переходов строк.
						--cnt_newlines = math.floor((#arr2 + 1) * 0.5) --Увеличиваем, но не сильно.
						cnt_newlines = 1
						--В конце добавляем переходы
						--Добавляем пустые строки в конец. TODO: Это надо только если предмет в инвентаре.
						while cnt_newlines > 0 do
							str = str .. "\n "
							cnt_newlines = cnt_newlines - 1
						end
					end
					--print("Check User Hint: "..str2)
					--Если первый раз, то отправляем запрос.
					if target ~= save_target or last_check_time + 1 < GetTime() then
						save_target = target
						last_check_time = GetTime()
						SendModRPCToServer(MOD_RPC.ShowMeHint.Hint, save_target.GUID, save_target)
					end
				else
					--print("target nil")
				end
				return old_SetString(text,str)
			end
		end)
	end
	
	--Обработчик на сервере
	AddModRPCHandler("ShowMeHint", "Hint", function(player, guid, item)
		if player.player_classified == nil then
			print("ERROR: player_classified not found!")
			return
		end
		if item ~= nil and item.components ~= nil then
			local s = GetTestString(item,player) --Формируем строку на сервере.
			if s ~= "" then
				player.player_classified.net_showme_hint:set(guid..";"..s) --Пакуем в строку и отсылаем обратно тому же игроку.
			end
		end
	end)

	--networking
	AddPrefabPostInit("player_classified",function(inst)
		inst.showme_hint = ""
		inst.net_showme_hint = _G.net_string(inst.GUID, "showme_hint", "showme_hint_dirty")
		if CLIENT_SIDE then
			inst:ListenForEvent("showme_hint_dirty",function(inst)
				inst.showme_hint = inst.net_showme_hint:value()
			end)
		end
	end)
end

--Обработка сундуков
do
	--[[
	If you want add your custom chest, use this code:
		TUNING.MONITOR_CHESTS = TUNING.MONITOR_CHESTS or {}
		TUNING.MONITOR_CHESTS.chestprefab = true
	--]]
	local MONITOR_CHESTS = { treasurechest=1, dragonflychest=1, skullchest=1, pandoraschest=1, minotaurchest=1,
		--bundle=1, --No container component. =\
		icebox=1, cookpot=1, -- No cookpot because it may be changed.
		chester=1, hutch=1,
		largechest=1, largeicebox=1, --Large Chest mod.
		safebox=1, safechest=1, safeicebox=1, --Safe mod.
		red_treasure_chest=1, purple_treasure_chest=1, green_treasure_chest=1, blue_treasure_chest=1, --Treasure Chests mod.
		backpack=1, candybag=1, icepack=1, piggyback=1, krampus_sack=1,
	}
	if TUNING.MONITOR_CHESTS then
		for k in pairs(TUNING.MONITOR_CHESTS) do
			MONITOR_CHESTS[k] = 1
		end
	--else
		--TUNING.MONITOR_CHESTS = MONITOR_CHESTS --Это не нужно.
	end
	local _active --Текущий предмет в курсоре (на клиенте).
	local _ing_prefab --Ингредиент. Через 5 секунд убирается.
	local net_string = _G.net_string
	local chests_around = {} --Массив всех сундуков в радиусе видимости клиента. Для хоста - все сундуки, но это норм.
	
	--[[
	_G.showme_count_chests = function() --debug function
		local cnt = 0
		for k,v in pairs(chests_around) do
			cnt = cnt + 1
		end
		print('Chests around:',cnt)
	end
	--]]
	
	local function OnClose(inst) --,err) --При закрытии сундука посылаем новые данные клиенту о его содержимом.
		local c = inst.components.container
		if not c then
			--[[if type(err) ~= "number" then err=nil end
			print('ERROR ShowMe: in ',inst.prefab,err)
			if not err then
				if inst.components then
					print("\tComponents:")
					for k in pairs(inst.components) do
						print("\t\t"..tostring(k))
					end
				else
					print("\tNo components at all!")
				end
			end
			if not err or err < 2000 then
				inst:DoTaskInTime(0,function(inst)
					OnClose(inst,err and (err+1) or 1)
				end)
			end--]]
			return
		end
		--if err then
		--	print("Found!!!!! Problem solved",err)
		--end
		if c:IsEmpty() then
			inst.net_ShowMe_chest:set('')
			return
		end
		local arr = {} -- [префаб]=true
		--[[ Отрывок из предыдущего сочинения (чтобы знать, что там происходит):
		if c.unwrappable and c.unwrappable.itemdata and type(c.unwrappable.itemdata) == 'table' then
			--По одной строке на каждый предмет.
			for i,v in ipairs(c.unwrappable.itemdata) do
				if v.prefab then
					--Пересылаем название префаба и количество дней.
					local delta = v.data and v.data.perishable and v.data.perishable.time
					local count = v.data and v.data.stackable and v.data.stackable.stack
					cn('perish_product', v.prefab, count or 0, delta and round2(delta/TUNING.TOTAL_DAY_TIME,1))
				end
			end
		end--]]
		for k,v in pairs(c.slots) do
			arr[tostring(v.prefab)] = true
			local u = v.components and v.components.unwrappable
			if u and u.itemdata then
				for i,v in ipairs(u.itemdata) do
					arr[v.prefab] = true --Добавляем префаб в упаковке.
				end
			end
		end
		local s
		for k in pairs(arr) do
			if s then
				s = s .. ' ' .. k --Только пробельные символы будут далее работать.
			else
				s = k
			end
		end
		inst.net_ShowMe_chest:set(s) --Посылаем данные.
	end
	
	--Обновляет подсветку сундука. Функция должна сама узнавать, что в руке игрока.
	local function UpdateChestColor(inst)
		local in_container = inst.ShowMe_chest_table and (
			(_active and inst.ShowMe_chest_table[_active.prefab])
			or (_ing_prefab and inst.ShowMe_chest_table[_ing_prefab])
		)
		if inst.b_ShowMe_changed_color then
			if not in_container then
				if inst.ShowMeColor then
					inst.ShowMeColor(true)
				else
					inst.AnimState:SetMultColour(1,1,1,1) --По умолчанию.
					inst.b_ShowMe_changed_color = nil
				end
			end
		else
			if in_container then
				if inst.ShowMeColor then
					inst.ShowMeColor(false)
				else
					inst.AnimState:SetMultColour(0.3,1,1,1)
					inst.b_ShowMe_changed_color = true
				end
			end
		end
	end
	
	local function OnShowMeChestDirty(inst)
		--inst.components.HuntGameLogic.hunt_kills = inst.components.HuntGameLogic.net_hunt_kills:value()
		local str = inst.net_ShowMe_chest:value()
		--inst.test_str = str --test
		--print('Test Chest:',str)
		local t = inst.ShowMe_chest_table
		for k in pairs(t) do
			t[k] = nil
		end
		for w in string.gmatch(str, "%S+") do
			t[w] = true
		end
		UpdateChestColor(inst) --Перерисовывает данный конкретный сундук, если изменилось его содержимое.
	end	

	local function InitChest(inst)
		inst.net_ShowMe_chest = net_string(inst.GUID, "ShowMe_chest", "ShowMe_chest_dirty" )
		if CLIENT_SIDE then
			inst:ListenForEvent("ShowMe_chest_dirty", OnShowMeChestDirty)
			chests_around[inst] = true
			inst.ShowMe_chest_table = {}
			inst.ShowTable = function() for k in pairs(inst.ShowMe_chest_table) do print(k) end end --debug
			inst:ListenForEvent('onremove', function(inst)
				chests_around[inst] = nil
			end)
		end
		if not SERVER_SIDE then
			return
		end
		inst:ListenForEvent("onclose", OnClose)
		inst:ListenForEvent("itemget", OnClose) --Для рюкзаков.
		inst:ListenForEvent("itemlose", OnClose)
		inst:DoTaskInTime(0,function(inst)
			OnClose(inst) --Изначально тоже посылаем данные, а не только при закрытии. Ведь сундук мог быть загружен.
		end)
	end
	
	for k in pairs(MONITOR_CHESTS) do
		AddPrefabPostInit(k,InitChest)
	end
	--Фиксим игрока, чтобы мониторить действия курсора.
	if CLIENT_SIDE then
		local function UpdateAllChestsAround()
			for k in pairs(chests_around) do
				UpdateChestColor(k)
			end
		end
		AddPrefabPostInit("inventory_classified",function(inst)
			inst:ListenForEvent("activedirty", function(inst)
				--print("ACTIVE:",inst._active:value())
				_active = inst._active:value()
				_ing_prefab = nil --Если взят предмет, то рецепт сразу же забываем.
				UpdateAllChestsAround() --Перерисовываем ВСЕ сундуки при каждом активном предмете или его отмене.
			end)
		end)
		
		local _ing_task
		local function UpdateIngredientView(player, prefab)
			_ing_prefab = prefab
			UpdateAllChestsAround()
			if _ing_task then
				_ing_task:Cancel()
			end
			_ing_task = player:DoTaskInTime(15,function(inst)
				_ing_prefab = nil
				_ing_task = nil
				UpdateAllChestsAround()
			end)
		end
		
		local ingredientui = _G.require 'widgets/ingredientui'
		local old_OnGainFocus = ingredientui.OnGainFocus

		function ingredientui:OnGainFocus(...)
			local prefab = self.ing and self.ing.texture and self.ing.texture:match('[^/]+$'):gsub('%.tex$', '')
			local player = self.parent and self.parent.parent and self.parent.parent.owner

			if prefab and player then
				--print("INGREDIENT:",prefab)
				UpdateIngredientView(player,prefab)
			end
			if old_OnGainFocus then
				return old_OnGainFocus(self, ...)
			end
		end
	end
end

------------------------------------------HOUND COMPANION---------------------------------------------------

local require = GLOBAL.require
local STRINGS = GLOBAL.STRINGS
local RECIPETABS = GLOBAL.RECIPETABS
local Recipe = GLOBAL.Recipe
local Ingredient = GLOBAL.Ingredient
local TECH = GLOBAL.TECH

GLOBAL.STRINGS.NAMES.KOKOCOLLAR = "Collar"
STRINGS.RECIPE_DESC.KOKOCOLLAR = "Call forth the hounds!"
STRINGS.CHARACTERS.GENERIC.DESCRIBE.KOKOCOLLAR= "A portal to a world unknown."

GLOBAL.STRINGS.NAMES.NIGHTMARETREAT = "Nightmare Treat"
STRINGS.RECIPE_DESC.NIGHTMARETREAT = "It moans softly."
STRINGS.CHARACTERS.GENERIC.DESCRIBE.NIGHTMARETREAT= "Concentrated evil in a bowl. \n Not very tasty."

GLOBAL.STRINGS.NAMES.DENMOTHERCOLLAR = "Heavy Collar"
STRINGS.RECIPE_DESC.DENMOTHERCOLLAR = "Moderately scarier."
STRINGS.CHARACTERS.GENERIC.DESCRIBE.DENMOTHERCOLLAR= "It emits a faint growling..."

STRINGS.NAMES.KOKO = "Koko"
STRINGS.CHARACTERS.GENERIC.DESCRIBE.KOKO = "Hello, friend!"

STRINGS.NAMES.LANA = "Lana"
STRINGS.CHARACTERS.GENERIC.DESCRIBE.LANA = "It's so adorable!"

STRINGS.NAMES.DENMOTHER = "Mother?"
STRINGS.CHARACTERS.GENERIC.DESCRIBE.DENMOTHER= "Why do I get the urge to call it mother?"

STRINGS.NAMES.DENGUARD = "Den Guard"
STRINGS.CHARACTERS.GENERIC.DESCRIBE.DENGUARD= "They don't seem to care I exist."

-- It's \n to start a new line in strings.

local Ingredient = GLOBAL.Ingredient
--AddRecipe("kokocollar", {Ingredient("glommerfuel", 1), Ingredient("transistor", 2), Ingredient("silk", 4)}, GLOBAL.RECIPETABS.MAGIC, GLOBAL.TECH.MAGIC_TWO, nil, nil, nil, nil, nil, "images/inventoryimages/kokocollar.xml", "kokocollar.tex" )
AddRecipe("nightmaretreat", {Ingredient("nightmarefuel", 4), Ingredient("monstermeat_dried", 3), Ingredient("berries", 5)}, GLOBAL.RECIPETABS.FARM, GLOBAL.TECH.MAGIC_TWO, nil, nil, nil, nil, nil, "images/inventoryimages/nightmaretreat.xml", "nightmaretreat.tex" )
--AddRecipe("denmothercollar", {Ingredient("nightmarefuel", 1), Ingredient("transistor", 2), Ingredient("silk", 4)}, GLOBAL.RECIPETABS.MAGIC, GLOBAL.TECH.MAGIC_THREE, nil, nil, nil, nil, nil, "images/inventoryimages/kokocollar.xml", "kokocollar.tex" )

---------------------------------------------DONT DROP------------------------------------------------------
local R_diao = GetModConfigData("rendiao")
local B_diao = GetModConfigData("baodiao")
local amu_diao = GetModConfigData("amudiao")
local zhuang_bei = GetModConfigData("zbdiao")
local R_d = R_diao - 3
local B_d = B_diao - 5
if R_d < 0 then R_d = 0 end if B_d < 0 then B_d = 0 end

AddComponentPostInit("container", function(Container, inst)
	function Container:DropSuiji(ondeath)
		local amu_x = true
		for k=1, self.numslots do
			local v = self.slots[k]
			--if amu_diao and amu_x and v and v.prefab == "amulet" then
				--amu_x = false
				--self:DropItem(v)
			--end
			--if B_diao ~= 0 and v and v.prefab == "reviver" then
				--self:DropItem(v)
			--end
		end
		for k=1, self.numslots do
			local v = self.slots[math.random(1, self.numslots)]
			if k > math.random(B_d, B_diao) then
				return false
			end
			--if v then
				--self:DropItem(v)
			--end
		end
	end
end)

AddComponentPostInit("inventory", function(Inventory, inst)
	Inventory.oldDropEverythingFn = Inventory.DropEverything
	function Inventory:DropSuiji(ondeath)
		local amu_x = true
		for k=1, self.maxslots do
			local v = self.itemslots[k]
			--if amu_diao and amu_x and v and v.prefab == "amulet" then
				--amu_x = false
				--self:DropItem(v, true, true)
			--end
			--if R_diao ~= 0 and v and v.prefab == "reviver" then
				--self:DropItem(v, true, true)
			--end
		end
		for k=1, self.maxslots do
			local v = self.itemslots[math.random(1, self.maxslots)]
			if k~=1 and k > math.random(R_d, R_diao) then
				return false
			end
			--if v then
				--self:DropItem(v, true, true)
			--end
		end
	end

	function Inventory:PlayerSiWang(ondeath)
		for k, v in pairs(self.equipslots) do
			if v:HasTag("backpack") and v.components.container then
				v.components.container:DropSuiji(true)
			end
		end
		if zhuang_bei then
			for k, v in pairs(self.equipslots) do
				--if not v:HasTag("backpack") then
					--self:DropItem(v, true, true)
				--end
			end
		end
		self.inst.components.inventory:DropSuiji(true)
	end

	function Inventory:DropEverything(ondeath, keepequip)
		if not inst:HasTag("player") then
			return Inventory:oldDropEverythingFn(ondeath, keepequip)
		else
			return Inventory:PlayerSiWang(ondeath)
		end
	end
end)
---------------------------------------------REALISTIC SPACING------------------------------------------------------
for _,v in pairs(GLOBAL.AllRecipes) do
   v.min_spacing = 1
end

AddPrefabPostInit("dug_berrybush", function(inst)
    if inst.components.deployable then
        inst.components.deployable:SetDeploySpacing(GLOBAL.DEPLOYSPACING.MEDIUM)
    end
end)
---------------------------------------------DURABILITY TOOLS/CLOTHES/WEAPON------------------------------------------------------
TUNING.HAMMER_LOOT_PERCENT = GetModConfigData("HAMMER_LOOT_PERCENT")
TUNING.AXE_USES = GetModConfigData("AXE_USES")
TUNING.HAMMER_USES = GetModConfigData("HAMMER_USES")
TUNING.SHOVEL_USES = GetModConfigData("SHOVEL_USES")
TUNING.PITCHFORK_USES = GetModConfigData("PITCHFORK_USES")
TUNING.PICKAXE_USES = GetModConfigData("PICKAXE_USES")
TUNING.BUGNET_USES = GetModConfigData("BUGNET_USES")
TUNING.FISHINGROD_USES = GetModConfigData("FISHINGROD_USES")
TUNING.MULTITOOL_AXE_PICKAXE_USES = GetModConfigData("MULTITOOL_AXE_PICKAXE_USES")

TUNING.SPEAR_USES = GetModConfigData("SPEAR_USES")
TUNING.SPIKE_USES = GetModConfigData("SPIKE_USES")
TUNING.BOOMERANG_USES = GetModConfigData("BOOMERANG_USES")
TUNING.NIGHTSWORD_USES = GetModConfigData("NIGHTSWORD_USES")
TUNING.HAMBAT_USES = GetModConfigData("HAMBAT_USES")
TUNING.BATBAT_USES = GetModConfigData("BATBAT_USES")
TUNING.RUINS_BAT_USES = GetModConfigData("RUINS_BAT_USES")
TUNING.WHIP_USES = GetModConfigData("TAIL_O_THREE_CATS_USES")
TUNING.WATHGRITHR_SPEAR_USES = GetModConfigData("WATHGRITHR_SPEAR_USES")

TUNING.MINERHAT_LIGHTTIME = TUNING.MINERHAT_LIGHTTIME * GetModConfigData("MINERHAT_LIGHTTIME")
TUNING.LANTERN_LIGHTTIME = TUNING.LANTERN_LIGHTTIME * GetModConfigData("LANTERN_LIGHTTIME")
TUNING.SPIDERHAT_PERISHTIME = TUNING.SPIDERHAT_PERISHTIME * GetModConfigData("SPIDERHAT_PERISHTIME")
TUNING.ONEMANBAND_PERISHTIME = TUNING.ONEMANBAND_PERISHTIME * GetModConfigData("ONEMANBAND_PERISHTIME")
TUNING.UMBRELLA_PERISHTIME = TUNING.UMBRELLA_PERISHTIME * GetModConfigData("UMBRELLA_PERISHTIME")
TUNING.EARMUFF_PERISHTIME = TUNING.EARMUFF_PERISHTIME * GetModConfigData("EARMUFF_PERISHTIME")
TUNING.WINTERHAT_PERISHTIME = TUNING.WINTERHAT_PERISHTIME * GetModConfigData("WINTERHAT_PERISHTIME")
TUNING.BEEFALOHAT_PERISHTIME = TUNING.BEEFALOHAT_PERISHTIME * GetModConfigData("BEEFALOHAT_PERISHTIME")
TUNING.TRUNKVEST_PERISHTIME = TUNING.TRUNKVEST_PERISHTIME * GetModConfigData("TRUNKVEST_PERISHTIME")
TUNING.SWEATERVEST_PERISHTIME = TUNING.SWEATERVEST_PERISHTIME * GetModConfigData("SWEATERVEST_PERISHTIME")
TUNING.HUNGERBELT_PERISHTIME = TUNING.HUNGERBELT_PERISHTIME * GetModConfigData("HUNGERBELT_PERISHTIME")
TUNING.WALRUSHAT_PERISHTIME = TUNING.WALRUSHAT_PERISHTIME * GetModConfigData("WALRUSHAT_PERISHTIME")
TUNING.FEATHERHAT_PERISHTIME = TUNING.FEATHERHAT_PERISHTIME * GetModConfigData("FEATHERHAT_PERISHTIME")
TUNING.TOPHAT_PERISHTIME = TUNING.TOPHAT_PERISHTIME * GetModConfigData("TOPHAT_PERISHTIME")
TUNING.STRAWHAT_PERISHTIME = TUNING.STRAWHAT_PERISHTIME * GetModConfigData("STRAWHAT_PERISHTIME")
TUNING.GRASS_UMBRELLA_PERISHTIME = TUNING.GRASS_UMBRELLA_PERISHTIME * GetModConfigData("GRASS_UMBRELLA_PERISHTIME")
TUNING.ICEHAT_PERISHTIME = TUNING.ICEHAT_PERISHTIME * GetModConfigData("ICEHAT_PERISHTIME")
TUNING.REFLECTIVEVEST_PERISHTIME = TUNING.REFLECTIVEVEST_PERISHTIME * GetModConfigData("REFLECTIVEVEST_PERISHTIME")
TUNING.HAWAIIANSHIRT_PERISHTIME = TUNING.HAWAIIANSHIRT_PERISHTIME * GetModConfigData("HAWAIIANSHIRT_PERISHTIME")
TUNING.CATCOONHAT_PERISHTIME = TUNING.CATCOONHAT_PERISHTIME * GetModConfigData("CATCOONHAT_PERISHTIME")
TUNING.RAINHAT_PERISHTIME = TUNING.RAINHAT_PERISHTIME * GetModConfigData("RAINHAT_PERISHTIME")
TUNING.RAINCOAT_PERISHTIME = TUNING.RAINCOAT_PERISHTIME * GetModConfigData("RAINCOAT_PERISHTIME")
TUNING.BEARGERVEST_PERISHTIME = TUNING.BEARGERVEST_PERISHTIME * GetModConfigData("BEARGERVEST_PERISHTIME")
TUNING.EYEBRELLA_PERISHTIME = TUNING.EYEBRELLA_PERISHTIME * GetModConfigData("EYEBRELLA_PERISHTIME")
TUNING.MOLEHAT_PERISHTIME = TUNING.MOLEHAT_PERISHTIME * GetModConfigData("MOLEHAT_PERISHTIME")
TUNING.ICESTAFF_USES = GetModConfigData("ICESTAFF_USES")
TUNING.FIRESTAFF_USES = GetModConfigData("FIRESTAFF_USES")
TUNING.YELLOWSTAFF_USES = GetModConfigData("YELLOWSTAFF_USES")
TUNING.ORANGESTAFF_USES = GetModConfigData("ORANGESTAFF_USES")
TUNING.GREENSTAFF_USES = GetModConfigData("GREENSTAFF_USES")
TUNING.TELESTAFF_USES = GetModConfigData("TELESTAFF_USES")
TUNING.REDAMULET_USES = GetModConfigData("REDAMULET_USES")
TUNING.BLUEAMULET_FUEL = TUNING.BLUEAMULET_FUEL * GetModConfigData("BLUEAMULET_FUEL")
TUNING.PURPLEAMULET_FUEL = TUNING.PURPLEAMULET_FUEL * GetModConfigData("PURPLEAMULET_FUEL")
TUNING.YELLOWAMULET_FUEL = TUNING.YELLOWAMULET_FUEL * GetModConfigData("YELLOWAMULET_FUEL")
TUNING.ORANGEAMULET_USES = GetModConfigData("ORANGEAMULET_USES")
TUNING.GREENAMULET_USES = GetModConfigData("GREENAMULET_USES")
TUNING.TORNADOSTAFF_USES = GetModConfigData("TORNADOSTAFF_USES")

---------------------------------------------AUTOTRAP------------------------------------------------------
local SpawnPrefab = GLOBAL.SpawnPrefab

local function ModifyTrapComponent(self)
    -- Remove the inventoryitem component before harvesting, re-add it and
    -- rearm the trap after.
    -- This assumes there are no additional component attributes set, which
    -- currently is the case for trap/birdtrap. In the future, either store/restore
    -- the component or its attributes, or limit it to specific prefabs and check
    -- for future changes.
    local Harvest_original = self.Harvest
    function self:Harvest(doer, ...)
        local was_inventoryitem = self.inst.components.inventoryitem ~= nil

        if was_inventoryitem then
            self.inst:RemoveComponent("inventoryitem")
        end

        Harvest_original(self, doer, ...)

        if self.inst:IsValid() and
           (self.inst.components.finiteuses == nil or
            self.inst.components.finiteuses:GetUses() > 0) then
            if was_inventoryitem then
                self.inst:AddComponent("inventoryitem")
            end

            -- Event handlers (e.g. in stategraphs/SGtrap.lua) will
            -- disarm the trap; thus schedule setting it.
            --~ self.inst:DoTaskInTime(2.0,
                                   --~ function()
                                       --~ -- TODO: Check if it has been picked up in the meantime
                                       --~ self:Set()
                                   --~ end)
            -- Or, assuming event order is kept and processed in a single thread,
            -- simply call the ondropped event handler, which should set it.
            self.inst:PushEvent("ondropped")
        end
    end

    -- Fix a bug which causes traps whose target became invalid
    -- to be uncheckable and prevents them from losing durability.
    local DoSpring_original = self.DoSpring
    function self:DoSpring(...)
        local return_value = DoSpring_original(self, ...)

        self.isset = false
        self.issprung = true

        return return_value
    end
end

AddComponentPostInit("trap", ModifyTrapComponent)


local function TrapComponentPostInit(self)
	function self:Explode(target)
		self:StopTesting()
		self.target = target
		self.issprung = true
		self.inactive = false
		if self.onexplode then
			self.onexplode(self.inst, target)
		end
		
		self.inst:DoPeriodicTask(4, function()
			if ( self.issprung ) then
				self:Reset()
			end
		end)
	end	
end

AddComponentPostInit("mine", TrapComponentPostInit)
-----------------------------------------------------------ICEBOX/STACK/STORAGE------------------------------------------------------------------

--ICEBOX
TUNING.PERISH_FRIDGE_MULT = GetModConfigData("speed")
--STACK 100
local require = GLOBAL.require
local stackable_replica = require "components/stackable_replica"
local IsServer = GLOBAL.TheNet:GetIsServer()

local size = GetModConfigData("MAXSTACKSIZE")
local TUNING = GLOBAL.TUNING
local net_byte = GLOBAL.net_byte

TUNING.STACK_SIZE_LARGEITEM = size
TUNING.STACK_SIZE_MEDITEM = size
TUNING.STACK_SIZE_SMALLITEM = size 

-- Define functions
local stackable_replica_ctorBase = stackable_replica._ctor or function() return true end    
function stackable_replica._ctor(self, inst)
    self.inst = inst

    self._stacksize = net_byte(inst.GUID, "stackable._stacksize", "stacksizedirty")
    self._maxsize = size
end

local stackable_replicaSetMaxSize_Base = stackable_replica.SetMaxSize or function() return true end
function stackable_replica:SetMaxSize(maxsize)
	self._maxsize = size
end

local stackable_replicaMaxSize_Base = stackable_replica.MaxSize or function() return true end
function stackable_replica:MaxSize()
	return self._maxsize
end

if IsServer then
AddPrefabPostInit("rabbit",function(inst)
   if(inst.components.stackable == nil) then
      inst:AddComponent("stackable")
   end
   inst.components.inventoryitem:SetOnDroppedFn(function(inst)
        inst.components.perishable:StopPerishing()
        inst.sg:GoToState("stunned")
        if inst.components.stackable then
    	    while inst.components.stackable:StackSize() > 1 do
    	        local item = inst.components.stackable:Get()
    	        if item then
    	            if item.components.inventoryitem then
    	                item.components.inventoryitem:OnDropped()
    	            end
    	            item.Physics:Teleport(inst.Transform:GetWorldPosition() )
    	        end
    	    end
    	 end
    end)
end)
end

-----------------------------------------------------------EXTRA SLOT------------------------------------------------------------------
local require = GLOBAL.require
local TheInput = GLOBAL.TheInput
local ThePlayer = GLOBAL.ThePlayer
local IsServer = GLOBAL.TheNet:GetIsServer()
local Inv = require "widgets/inventorybar"


-- for key,value in pairs(GLOBAL.EQUIPSLOTS) do print('4r',key,value) end

GLOBAL.EQUIPSLOTS=
{
    HANDS = "hands",
    HEAD = "head",
    BODY = "body",
    BACK = "back",
    NECK = "neck",
}
GLOBAL.EQUIPSLOT_IDS = {}
local slot = 0
for k, v in pairs(GLOBAL.EQUIPSLOTS) do
    slot = slot + 1
    GLOBAL.EQUIPSLOT_IDS[v] = slot
end
slot = nil

AddComponentPostInit("resurrectable", function(self, inst)
    local original_FindClosestResurrector = self.FindClosestResurrector
    local original_CanResurrect = self.CanResurrect
    local original_DoResurrect = self.DoResurrect

    self.FindClosestResurrector = function(self)
        if IsServer and self.inst.components.inventory then
            local item = self.inst.components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.NECK)
            if item and item.prefab == "amulet" then
                return item
            end
        end
        original_FindClosestResurrector(self)
    end

    self.CanResurrect = function(self)
        if IsServer and self.inst.components.inventory then
            local item = self.inst.components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.NECK)
            if item and item.prefab == "amulet" then
                return true
            end
        end
        original_CanResurrect(self)
    end

    self.DoResurrect = function(self)
        self.inst:PushEvent("resurrect")
        if IsServer and self.inst.components.inventory then
            local item = self.inst.components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.NECK)
            if item and item.prefab == "amulet" then
                self.inst.sg:GoToState("amulet_rebirth")
                return true
            end
        end
        original_DoResurrect(self)
    end
end)

AddComponentPostInit("inventory", function(self, inst)
    local original_Equip = self.Equip
    self.Equip = function(self, item, old_to_active)
        if original_Equip(self, item, old_to_active) and item and item.components and item.components.equippable then
            local eslot = item.components.equippable.equipslot
            if self.equipslots[eslot] ~= item then
                if eslot == GLOBAL.EQUIPSLOTS.BACK and item.components.container ~= nil then
                    self.inst:PushEvent("setoverflow", { overflow = item })
                end
            end
            return true
        else
            return
        end
    end

    self.GetOverflowContainer = function()
        if self.ignoreoverflow then
            return
        end
        local item = self:GetEquippedItem(GLOBAL.EQUIPSLOTS.BACK)
        return item ~= nil and item.components.container or nil
    end
end)

AddGlobalClassPostConstruct("widgets/inventorybar", "Inv", function()
    local Inv_Refresh_base = Inv.Refresh or function() return "" end
    local Inv_Rebuild_base = Inv.Rebuild or function() return "" end

    function Inv:LoadExtraSlots(self)
        self.bg:SetScale(1.35,1,1.25)
        self.bgcover:SetScale(1.35,1,1.25)

        if self.addextraslots == nil then
            self.addextraslots = 1

            self:AddEquipSlot(GLOBAL.EQUIPSLOTS.BACK, "images/back.xml", "back.tex")
            self:AddEquipSlot(GLOBAL.EQUIPSLOTS.NECK, "images/neck.xml", "neck.tex")
        -- else
            -- GLOBAL.GetPlayer().HUD.controls.stickyrecipepopup:Refresh()

            if self.inspectcontrol then
                local W = 68
                local SEP = 12
                local INTERSEP = 28
                local inventory = self.owner.replica.inventory
                local num_slots = inventory:GetNumSlots()
                local num_equip = #self.equipslotinfo
                local num_buttons = self.controller_build and 0 or 1
                local num_slotintersep = math.ceil(num_slots / 5)
                local num_equipintersep = num_buttons > 0 and 1 or 0
                local total_w = (num_slots + num_equip + num_buttons) * W + (num_slots + num_equip + num_buttons - num_slotintersep - num_equipintersep - 1) * SEP + (num_slotintersep + num_equipintersep) * INTERSEP
            	self.inspectcontrol.icon:SetPosition(-4, 6)
            	self.inspectcontrol:SetPosition((total_w - W) * .5 + 3, -6, 0)
            end
        end

        --if not self.controller_build then
        --    self.bg:SetScale(1.22, 1, 1)
        --    self.bgcover:SetScale(1.22, 1, 1)

        --    self.inspectcontrol = self.root:AddChild(TEMPLATES.IconButton(atlas_name, image_name, STRINGS.UI.HUD.INSPECT_SELF, false, false, function() self.owner.HUD:InspectSelf() end, {size = 40}, "self_inspect_mod.tex"))
        --    self.inspectcontrol.icon:SetScale(.7)
        --    self.inspectcontrol.icon:SetPosition(-4, 6)
        --    self.inspectcontrol:SetScale(1.25)
        --    self.inspectcontrol:SetPosition((total_w - W) * .5 + 3, -6, 0)
        --else
        --    self.bg:SetScale(1.15, 1, 1)
        --    self.bgcover:SetScale(1.15, 1, 1)

        --    if self.inspectcontrol ~= nil then
        --        self.inspectcontrol:Kill()
        --        self.inspectcontrol = nil
        --    end
        --end
    end

    function Inv:Refresh()
        Inv_Refresh_base(self)
        Inv:LoadExtraSlots(self)
    end

    function Inv:Rebuild()
        Inv_Rebuild_base(self)
        Inv:LoadExtraSlots(self)
    end
end)

AddPrefabPostInit("inventory_classified", function(inst)
    function GetOverflowContainer(inst)
        local item = inst.GetEquippedItem(inst, GLOBAL.EQUIPSLOTS.BACK)
        return item ~= nil and item.replica.container or nil
    end

    function Count(item)
        return item.replica.stackable ~= nil and item.replica.stackable:StackSize() or 1
    end

    function Has(inst, prefab, amount)
        local count =
            inst._activeitem ~= nil and
            inst._activeitem.prefab == prefab and
            Count(inst._activeitem) or 0

        if inst._itemspreview ~= nil then
            for i, v in ipairs(inst._items) do
                local item = inst._itemspreview[i]
                if item ~= nil and item.prefab == prefab then
                    count = count + Count(item)
                end
            end
        else
            for i, v in ipairs(inst._items) do
                local item = v:value()
                if item ~= nil and item ~= inst._activeitem and item.prefab == prefab then
                    count = count + Count(item)
                end
            end
        end

        local overflow = GetOverflowContainer(inst)
        if overflow ~= nil then
            local overflowhas, overflowcount = overflow:Has(prefab, amount)
            count = count + overflowcount
        end

        return count >= amount, count
    end

    -- function IsBusy(inst)
    --     return inst._busy or inst._parent == nil
    -- end

    -- local PushItemGet = inst.PushItemGet
    -- local PushStackSize = inst.PushStackSize
    -- local ConsumeByName = inst.ConsumeByName

    -- function UseItemFromInvTile(inst, item)
    --     if not IsBusy(inst) and
    --         inst._parent ~= nil and
    --         inst._parent.components.playeractionpicker ~= nil and
    --         inst._parent.components.playercontroller ~= nil then
    --         local actions = inst._activeitem ~= nil and
    --             inst._parent.components.playeractionpicker:GetUseItemActions(item, inst._activeitem, true) or
    --             inst._parent.components.playeractionpicker:GetInventoryActions(item)
    --         if #actions > 0 then
    --             if actions[1].action == GLOBAL.ACTIONS.RUMMAGE then
    --                 local overflow = GetOverflowContainer(inst)
    --                 if overflow ~= nil and overflow.inst == item then
    --                     if overflow:IsOpenedBy(inst._parent) then
    --                         overflow:Close()
    --                     else
    --                         overflow:Open(inst._parent)
    --                     end
    --                     return
    --                 end
    --             end
    --             inst._parent.components.playercontroller:RemoteUseItemFromInvTile(actions[1], item)
    --         end
    --     end
    -- end

    -- function ReceiveItem(inst, item, count)
    --     if IsBusy(inst) then
    --         return
    --     end
    --     local overflow = GetOverflowContainer(inst)
    --     overflow = overflow and overflow.classified or nil
    --     if overflow ~= nil and overflow:IsBusy() then
    --         return
    --     end
    --     local isstackable = item.replica.stackable ~= nil
    --     local originalstacksize = isstackable and item.replica.stackable:StackSize() or 1
    --     if not isstackable or inst._parent.replica.inventory == nil or not inst._parent.replica.inventory:AcceptsStacks() then
    --         for i, v in ipairs(inst._items) do
    --             if v:value() == nil then
    --                 local giveitem = SlotItem(item, i)
    --                 PushItemGet(inst, giveitem)
    --                 if originalstacksize > 1 then
    --                     PushStackSize(inst, item, nil, nil, 1, false, true)
    --                     return originalstacksize - 1
    --                 else
    --                     return 0
    --                 end
    --             end
    --         end
    --         if overflow ~= nil then
    --             return overflow:ReceiveItem(item, count)
    --         end
    --     else
    --         local originalcount = count and math.min(count, originalstacksize) or originalstacksize
    --         count = originalcount
    --         if item.replica.equippable ~= nil then
    --             local eslot = item.replica.equippable:EquipSlot()
    --             local equip = inst:GetEquippedItem(eslot)
    --             if equip ~= nil and
    --                 equip.prefab == item.prefab and
    --                 equip.replica.stackable ~= nil and
    --                 not equip.replica.stackable:IsFull() then
    --                 local stacksize = equip.replica.stackable:StackSize() + count
    --                 local maxsize = equip.replica.stackable:MaxSize()
    --                 if stacksize > maxsize then
    --                     count = math.max(stacksize - maxsize, 0)
    --                     stacksize = maxsize
    --                 else
    --                     count = 0
    --                 end
    --                 PushStackSize(inst, equip, stacksize, true, nil, nil, nil, SlotEquip(equip, eslot))
    --             end
    --         end
    --         if count > 0 then
    --             local emptyslot = nil
    --             for i, v in ipairs(inst._items) do
    --                 local slotitem = v:value()
    --                 if slotitem == nil then
    --                     if emptyslot == nil then
    --                         emptyslot = i
    --                     end
    --                 elseif slotitem.prefab == item.prefab and
    --                     slotitem.replica.stackable ~= nil and
    --                     not slotitem.replica.stackable:IsFull() then
    --                     local stacksize = slotitem.replica.stackable:StackSize() + count
    --                     local maxsize = slotitem.replica.stackable:MaxSize()
    --                     if stacksize > maxsize then
    --                         count = math.max(stacksize - maxsize, 0)
    --                         stacksize = maxsize
    --                     else
    --                         count = 0
    --                     end
    --                     PushStackSize(inst, slotitem, stacksize, true, nil, nil, nil, SlotItem(slotitem, i))
    --                     if count <= 0 then
    --                         break
    --                     end
    --                 end
    --             end
    --             if count > 0 then
    --                 if emptyslot ~= nil then
    --                     local giveitem = SlotItem(item, emptyslot)
    --                     PushItemGet(inst, giveitem)
    --                     if count ~= originalstacksize then
    --                         PushStackSize(inst, item, nil, nil, count, false, true)
    --                     end
    --                     count = 0
    --                 elseif overflow ~= nil then
    --                     local remainder = overflow:ReceiveItem(item, count)
    --                     if remainder ~= nil then
    --                         count = math.max(count - (originalstacksize - remainder), 0)
    --                     end
    --                 end
    --             end
    --         end
    --         if count ~= originalcount then
    --             return originalstacksize - (originalcount - count)
    --         end
    --     end
    -- end

    -- function RemoveIngredients(inst, recipe, ingredientmod)
    --     if IsBusy(inst) then
    --         return false
    --     end
    --     local overflow = GetOverflowContainer(inst)
    --     overflow = overflow and overflow.classified or nil
    --     if overflow ~= nil and overflow:IsBusy() then
    --         return false
    --     end
    --     for i, v in ipairs(recipe.ingredients) do
    --         local amt = math.max(1, GLOBAL.RoundBiasedUp(v.amount * ingredientmod))
    --         inst.ConsumeByName(inst, v.type, amt, overflow)
    --     end
    --     return true
    -- end

    if not IsServer then
        inst.GetOverflowContainer = GetOverflowContainer
        inst.Has = Has
        -- inst.UseItemFromInvTile = UseItemFromInvTile
        -- inst.ReceiveItem = ReceiveItem
        -- inst.RemoveIngredients = RemoveIngredients
    end
end)

AddStategraphPostInit("wilson", function(self)
    for key,value in pairs(self.states) do
        if value.name == 'amulet_rebirth' then
            local original_amulet_rebirth_onexit = self.states[key].onexit


            self.states[key].onexit = function(inst)
                local item = inst.components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.NECK)
                if item and item.prefab == "amulet" then
                    item = inst.components.inventory:RemoveItem(item)
                    if item then
                        item:Remove()
                        item.persists = false
                    end
                end
                original_amulet_rebirth_onexit(inst)
            end
        end
    end
end)

function backpackpostinit(inst)
    if IsServer then
        inst.components.equippable.equipslot = GLOBAL.EQUIPSLOTS.BACK or GLOBAL.EQUIPSLOTS.BODY
    end
end

function amuletpostinit(inst)
    if IsServer then
        inst.components.equippable.equipslot = GLOBAL.EQUIPSLOTS.NECK or GLOBAL.EQUIPSLOTS.BODY
    end
end


AddPrefabPostInit("amulet", amuletpostinit)
AddPrefabPostInit("blueamulet", amuletpostinit)
AddPrefabPostInit("purpleamulet", amuletpostinit)
AddPrefabPostInit("orangeamulet", amuletpostinit)
AddPrefabPostInit("greenamulet", amuletpostinit)
AddPrefabPostInit("yellowamulet", amuletpostinit)

AddPrefabPostInit("backpack", backpackpostinit)
AddPrefabPostInit("krampus_sack", backpackpostinit)
AddPrefabPostInit("piggyback", backpackpostinit)
AddPrefabPostInit("icepack", backpackpostinit)


------------------------------------------------------NOVOTE---------------------------------------------------
local _G = GLOBAL
local TheNet = _G.TheNet

local UserCommands = _G.require("usercommands")
local VoteUtil = _G.require("voteutil")

local no_rollback = GetModConfigData("no_rollback")
local no_regenerate = GetModConfigData("no_regenerate")
local no_kick = GetModConfigData("no_kick")
local kick_without = GetModConfigData("kick_without")

local show_starter = GetModConfigData("show_starter")

local reflect_kick = GetModConfigData("reflect_kick")

local kick_without_cache = {}
for key, value in pairs(kick_without) do
	kick_without_cache[value] = true
end


function showstarter(src, data)
	local cmd = UserCommands.GetCommandFromHash(data.commandhash)
	if cmd ~= nil then
		print("[NoVote]Vote Start by:", data.starteruserid, cmd.name, data.commandhash, data.targetuserid)
		if kick_without_cache[data.targetuserid] ~= nil and cmd.name == "kick" then
			print("[NoVote]", data.starteruserid, "try to kick", data.targetuserid)
			if reflect_kick then
			    TheNet:Kick(data.starteruserid, nil)
			end
		end
	end
end

function WaitActivated(inst)

if TheNet and TheNet:GetIsServer() then -- Is server
print("[NoVote]init...")

print("[NoVote]AddUserCommand", AddUserCommand)
print("[NoVote]_G.AddUserCommand", _G.AddUserCommand)
print("[NoVote]AddModUserCommand", AddModUserCommand)
print("[NoVote]_G.AddModUserCommand", _G.AddModUserCommand)
print("[NoVote]AddVoteCommand", AddVoteCommand)

--print("[NoVote]smallhash", smallhash)
print("[NoVote]_G.smallhash", _G.smallhash)
--print("[NoVote]GetCommandFromName", GetCommandFromName)
print("[NoVote]UserCommands.GetCommandFromName", UserCommands.GetCommandFromName)

	if show_starter then
		inst:ListenForEvent("ms_startvote", showstarter, inst)
	end

	if no_rollback then
print("[NoVote]no_rollback", no_rollback)
print("[NoVote]rollback Cmd", UserCommands.GetCommandFromName("rollback"))
		local rollbackCmd = UserCommands.GetCommandFromName("rollback")
		if rollbackCmd ~= nil then
			rollbackCmd.votetimeout = 60
--			rollbackCmd.voteminpasscount = nil
--			rollbackCmd.votecountvisible = true
--			rollbackCmd.voteallownotvoted = true
--			rollbackCmd.votecanstartfn = VoteUtil.DefaultCanStartVote
--			rollbackCmd.voteresultfn = VoteUtil.YesNoUnanimousVote
			rollbackCmd.serverfn = function(params, caller)
				if caller ~= nil then
					print("[NoVote][rollback]serverfn", caller, caller.name, caller.userid)
				else
					print("[NoVote][rollback]serverfn", caller)
				end
			end
		else
			print("[NoVote][rollback]Patch Cmd Faild!!")
		end
print("[NoVote]rollback2", UserCommands.GetCommandFromName("rollback"))
	end

	if no_regenerate then
print("[NoVote]no_regenerate", no_regenerate)
print("[NoVote]regenerate", UserCommands.GetCommandFromName("regenerate"))
		local regenerateCmd = UserCommands.GetCommandFromName("regenerate")
		if regenerateCmd ~= nil then
			regenerateCmd.votetimeout = 60
--			regenerateCmd.voteminpasscount = nil
--			regenerateCmd.votecountvisible = true
--			regenerateCmd.voteallownotvoted = true
--			regenerateCmd.votecanstartfn = VoteUtil.DefaultCanStartVote
--			regenerateCmd.voteresultfn = VoteUtil.YesNoUnanimousVote
			regenerateCmd.serverfn = function(params, caller)
				if caller ~= nil then
					print("[NoVote][regenerate]serverfn", caller, caller.name, caller.userid)
				else
					print("[NoVote][regenerate]serverfn", caller)
				end
			end
		else
			print("[NoVote][regenerate]Patch Cmd Faild!!")
		end
print("[NoVote]regenerate2", UserCommands.GetCommandFromName("regenerate"))
	end

	if no_kick or (#kick_without ~= 0) then
print("[NoVote]no_kick", no_kick)
print("[NoVote]kick_without", #kick_without, kick_without)
print("[NoVote]kick", UserCommands.GetCommandFromName("kick"))

		local kickCmd = UserCommands.GetCommandFromName("kick")
		if kickCmd ~= nil then
			kickCmd.votetimeout = 90
--			kickCmd.voteminpasscount = 3
--			kickCmd.votecountvisible = true
--			kickCmd.voteallownotvoted = true
--			kickCmd.votecanstartfn = VoteUtil.DefaultCanStartVote
--			kickCmd.voteresultfn = VoteUtil.YesNoMajorityVote
			kickCmd.localfn = function(params, caller)
				if caller ~= nil then
					print("[NoVote][kick]serverfn", caller, caller.name, caller.userid)
				else
					print("[NoVote][kick]serverfn", caller)
				end

				if no_kick then
					return
				end

				--NOTE: must support nil caller for voting
				local clientid = _G.UserToClientID(params.user)
				if clientid ~= nil then
					if kick_without_cache[clientid] == nil then
					    TheNet:Kick(clientid, caller == nil and _G.TUNING.VOTE_KICK_TIME or nil)
					else
						print("[NoVote][kick]can't kick", params.user, clientid)
					end
				end
			end
		else
			print("[NoVote][kick]Patch Cmd Faild!!")
		end
print("[NoVote]kick2", UserCommands.GetCommandFromName("kick"))
	end
end


end

AddPrefabPostInit("world", WaitActivated)

------------------------------------------------------STARTING ITEMS---------------------------------------------------
local OnPlayerSpawn = function(src, player)
    player.prev_OnNewSpawn = player.OnNewSpawn
    player.OnNewSpawn = function()
   
        local start_items = {}
        
        --for season starting items
        if GetModConfigData("is_season_items") == 1 then
            --for spring
            if GLOBAL.TheWorld.state.isspring or (GLOBAL.TheWorld.state.iswinter and GLOBAL.TheWorld.state.remainingdaysinseason < 3) then
                table.insert(start_items, "umbrella")
            end
            --for summer
            if GLOBAL.TheWorld.state.issummer or (GLOBAL.TheWorld.state.isspring and GLOBAL.TheWorld.state.remainingdaysinseason < 3) then
                table.insert(start_items, "heatrock")
                table.insert(start_items, "reflectivevest")
		table.insert(start_items, "umbrella")
                for i = 1, GetModConfigData("food_give") do table.insert(start_items, "meatballs") end
        	for i = 1, GetModConfigData("num_nitre") do table.insert(start_items, "nitre") end
            end
            --for winter
            if GLOBAL.TheWorld.state.iswinter or (GLOBAL.TheWorld.state.isautumn and GLOBAL.TheWorld.state.remainingdaysinseason < 3) then
    			table.insert(start_items, "heatrock")
                table.insert(start_items, "winterhat")
                for i = 1, GetModConfigData("food_give") do table.insert(start_items, "meatballs") end
            end
        end

        --for all season
        for i = 1, GetModConfigData("num_goldnugget") do table.insert(start_items, "goldnugget") end
        for i = 1, GetModConfigData("num_flint") do table.insert(start_items, "flint") end
        for i = 1, GetModConfigData("num_rocks") do table.insert(start_items, "rocks") end
        for i = 1, GetModConfigData("num_log") do table.insert(start_items, "log") end
        for i = 1, GetModConfigData("num_cutgrass") do table.insert(start_items, "cutgrass") end
        for i = 1, GetModConfigData("num_twigs") do table.insert(start_items, "twigs") end
 	--for i = 1, GetModConfigData("is_season_items") do table.insert(start_items, "goldenshovel") end

        for k,v in pairs(start_items) do
            local item = GLOBAL.SpawnPrefab(v)
            
            player.components.inventory:GiveItem(item)
        end   
         
        if player.prev_OnNewSpawn ~= nil then
            player:prev_OnNewSpawn()
            player.prev_OnNewSpawn = nil
        end
    end
end


local function ListenForPlayers(inst)
    if GLOBAL.TheWorld.ismastersim then
        inst:ListenForEvent("ms_playerspawn", OnPlayerSpawn)
    end
end

AddPrefabPostInit("world", ListenForPlayers)

------------------------------------------------------DIGABLE REED---------------------------------------------------
--[[		Created by: Ysovuka/Kzisor		]]

-- Import the lib use.
modimport("libs/use.lua")

-- Import the mod environment as our environment.
env = use "libs/mod_env"

-- Create the string name for dug reeds.
STRINGS.NAMES.DUG_REEDS = "Dug Reeds"

-- Import the prefab files.


-- Import data/reeds/base.lua
use "data/prefabs/reeds"

------------------------------------------------------NO DISEASE/GEKKO---------------------------------------------------
local modmastersim = GLOBAL.TheNet:GetIsMasterSimulation()

local SpawnPrefab = GLOBAL.SpawnPrefab
local TUNING = GLOBAL.TUNING

-- no more grass morphing
TUNING.GRASSGEKKO_MORPH_CHANCE = 0

-- no more disease appearing
TUNING.DISEASE_CHANCE = 0
TUNING.DISEASE_DELAY_TIME = 0
TUNING.DISEASE_DELAY_TIME_VARIANCE = 0

if modmastersim then
	-- gekkos into grass
	local function TurnIntoGrass(inst)
		local x, y, z = inst.Transform:GetWorldPosition()
		local grass = SpawnPrefab("grass")
		grass.Transform:SetPosition(x, y, z)
		inst:Remove()
	end
	local function DelaySwap(inst)
		inst:DoTaskInTime(0, TurnIntoGrass)
	end
	AddPrefabPostInit("grassgekko", DelaySwap)

	-- cure stuff if there is something diseased
	local function TurnIntoNormal(inst)
		local x, y, z = inst.Transform:GetWorldPosition()
		local normal = SpawnPrefab(inst.prefab)
		normal.Transform:SetPosition(x, y, z)
		inst:Remove()
	end
	local function DelayCure(self)
		if self:IsDiseased() or self:IsBecomingDiseased() then
			self.inst:DoTaskInTime(0, TurnIntoNormal)
		end
	end
	AddComponentPostInit("diseaseable", DelayCure)
end

------------------------------------------------------WORLD REGENERATION---------------------------------------------------
local assert = GLOBAL.assert

-- Server Console Commands

-- for i = 1,16 do TheWorld.components.ResourceRespawnerComponent:ClockChanged() end
-- Above command advances the world regeneration component by 1 day, use it for debug purposes

-- c_spawn('mound').components.RenewableResourceComponent.shouldAdd = true
-- Above command spawns a prefab, mound in this case, with the respawnable component

-- c_find('mound').components.RenewableResourceComponent.shouldAdd = false; c_find('mound'):Remove()
-- Above command find a nearest prefab with the respawnable component and disables it so it can be removed and it does not spawn back

-- TheWorld.components.ResourceRespawnerComponent.turfs=nil; TheWorld.components.ResourceRespawnerComponent:InitTurfs()
-- Above command reinitializes saved turfs to the current world state; EnableTurfsRegeneration option has to be enabled for it to work


--
-- load & check configuration options
--

-- get configuration option that should be number
-- @param optionname string  name of the option
-- @param min number  minimum value
--            nil  don't check minimum
-- @param max number  maximum value
--            nil  don't check maximum
-- @param defaulvalue number  default value if raw value is invalid
-- @param forcelocal  force reading from locally saved options, not server's
-- @return number
local function GetValidatedModConfigNumber(optionname, min, max, defaultvalue, forcelocal)
	assert(type(optionname) == "string", string.format("Parameter optionname '%s' has type %s, expected string.", tostring(optionname), type(optionname)))
	assert(min == nil or type(min) == "number", string.format("Parameter min '%s' has type %s, expected nil or number.", tostring(min), type(min)))
	assert(max == nil or type(max) == "number", string.format("Parameter max '%s' has type %s, expected nil or number.", tostring(max), type(max)))
	assert(type(defaultvalue) == "number", string.format("Parameter defaultvalue '%s' has type %s, expected number.", tostring(defaultvalue), type(defaultvalue)))

	local rawvalue = GetModConfigData(optionname, forcelocal)
	local val = GLOBAL.tonumber(rawvalue)
	if val == nil then
		moderror(string.format("Value of configuration option %s '%s' is invalid, setting to %s.", optionname, tostring(rawvalue), tostring(defaultvalue)))
		val = defaultvalue
	end
	local val2 = min == nil and val or (val < min and min or val)
	val2 = max == nil and val2 or (val2 > max and max or val2)
	if val2 ~= val then
		print(string.format("Value of configuration option %s '%s' is out of limits, setting to %s.", optionname, tostring(val), tostring(val2)))
	end

	return val2
end

-- get configuration option that should be boolean
-- @param optionname string  name of the option
-- @param forcelocal  force reading from locally saved options, not server's
-- @return boolean
local function GetValidatedModConfigBoolean(optionname, forcelocal)
	assert(type(optionname) == "string", string.format("Parameter optionname '%s' has type %s, expected string.", tostring(optionname), type(optionname)))

	local rawvalue = GetModConfigData(optionname, forcelocal)
	local val = rawvalue == true or rawvalue == "true" or rawvalue == 1 or rawvalue == "1"

	return val
end

local WorldResourceRegenerationSpeed = GetValidatedModConfigNumber("WorldResourceRegenerationSpeed", 0, 100, 5)
local SpawnFreeZone = GetValidatedModConfigNumber("SpawnFreeZone", 0, 100, 4)
local EnableGraveStoneRemoval = GetValidatedModConfigNumber("EnableGraveStoneRemoval", 0, 1, 0)
local EnableDebug = GetValidatedModConfigNumber("EnableDebug", 0, 1, 0)
local EnableTurfsRegeneration = GetValidatedModConfigBoolean("EnableTurfsRegeneration")
local CustomPrefabs = GetModConfigData("CustomPrefabs")
if type(CustomPrefabs) ~= "string" then
	moderror(string.format("Value of configuration option CustomPrefabs '%s' is invalid, setting to ''.", tostring(CustomPrefabs)))
	CustomPrefabs = ""
else
	CustomPrefabs = GLOBAL.TrimString(CustomPrefabs)
end

if EnableDebug > 0 then
	print(string.format("[WorldResourceRegeneration] WorldResourceRegenerationSpeed=%s, SpawnFreeZone=%s, EnableGraveStoneRemoval=%s, EnableDebug=%s, EnableTurfsRegeneration=%s, CustomPrefabs='%s'",
		tostring(WorldResourceRegenerationSpeed), tostring(SpawnFreeZone), tostring(EnableGraveStoneRemoval), tostring(EnableDebug), tostring(EnableTurfsRegeneration), tostring(CustomPrefabs)
	))
end

-- process CustomPrefabs configuration option
local customprefabs = {} -- associative array, item is prefabname=speed
if GLOBAL.TheNet:GetIsServer() and CustomPrefabs ~= "" then
	-- use string.split from util.lua; now that original ModeratorCommands mod is incompatible with the game, its overwriting of string.split shouldn't cause problems
	local list = string.split(CustomPrefabs, ",")
	for _, prefabspec in ipairs(list) do
		local speclist = string.split(prefabspec, ":")
		local prefabname = GLOBAL.TrimString(speclist[1] or "")
		if prefabname == "" then
			moderror(string.format("[WorldResourceRegeneration] Prefab name of custom prefab specification '%s' is empty.", tostring(prefabspec)))
		elseif customprefabs[prefabname] ~= nil then
			-- duplicate specification
			print(string.format("[WorldResourceRegeneration] Encountered duplicate custom prefab specification '%s', previous was '%s:%s', ignoring it.",
				tostring(prefabspec), tostring(prefabname), tostring(customprefabs[prefabname])
			))
		else
			-- the check, whether the prefab name is actually registered, can't be done now, do it later
			local speedstr = GLOBAL.TrimString(speclist[2] or "")
			local speed = GLOBAL.tonumber(speedstr)
			if speed == nil then
				moderror(string.format("[WorldResourceRegeneration] Speed in custom prefab specification '%s' is invalid, ignoring it.", tostring(prefabspec)))
			else
				local speed2 = math.clamp(math.ceil(speed), 0, math.ceil(2*WorldResourceRegenerationSpeed - 1))
				if speed2 ~= speed then
					print(string.format("[WorldResourceRegeneration] Speed '%s' in custom prefab specification '%s' is out of limits, using %s.",
						tostring(speed), tostring(prefabspec), tostring(speed2)
					))
					speed = speed2
				end
				customprefabs[prefabname] = speed
			end
		end
	end

	AddSimPostInit(function()
		-- check whether prefab names from custom prefabs options are actually registered
		for prefabname, speed in pairs(customprefabs) do
			if GLOBAL.Prefabs[prefabname] == nil then
				-- prefab not found, don't error out, just print a warning
				-- so it's possible to have specified prefabs from mods that are temporarily disabled without needing to edit custom prefabs configuration option
				print(string.format("[WorldResourceRegeneration] Prefab '%s' from custom prefab specification '%s:%s' is not registered.",
					tostring(prefabname), tostring(prefabname), tostring(speed)
				))
			end
		end
	end)
end


-- Regenerate World Resources
if WorldResourceRegenerationSpeed ~= 0 then
	local renew_1day =	{
		"fireflies",
		-- 3rd Party mods
	}
	for k, prefab in ipairs(renew_1day) do -- Very fast respawning items (next day)
		if customprefabs[prefab] == nil then -- skip custom prefabs
			AddPrefabPostInit(prefab, function(inst)
				if GLOBAL.TheWorld.ismastersim then
					inst:AddComponent("RenewableResourceComponent")
					inst.components.RenewableResourceComponent.delay = 1
				end
			end)
		end
	end

	local renew_short =	{
		"berrybush",
--		"carrot_planted", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"deciduoustrees", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"evergreens", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"flint", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"flower", -- Configurable in the worldgenoverride.lua or already renewable in game
		"grass",
		"rock1",
		"rock2",
		"sapling",
		-- 3rd Party mods
	}
	for k, prefab in ipairs(renew_short) do -- Items lying on the ground that are frequently harvested and very basic resources that should always be available
		if customprefabs[prefab] == nil then -- skip custom prefabs
			AddPrefabPostInit(prefab, function(inst)
				if GLOBAL.TheWorld.ismastersim then
					inst:AddComponent("RenewableResourceComponent")
					inst.components.RenewableResourceComponent.delay = math.ceil(0.5*WorldResourceRegenerationSpeed)
				end
			end)
		end
	end

	local renew_normal = {
		"beefalo",
		"beehive",
		"bishop",
--		"blueprint", -- No renewal, coordinates lost
		"cactus",
		"catcoonden",
		"cave_banana_tree",
		"cave_fern",
		"chessjunk1",
		"chessjunk2",
		"chessjunk3",
--		"flower_cave", -- Configurable in the worldgenoverride.lua or already renewable in game
		"flower_evil",
		"knight",
		"lichen",
		"livingtree",
		"mandrake_planted",
		"marblepillar",
		"marbletree",
		"marsh_bush",
		"marsh_tree",
		"mermhouse",
		"minotaur",
		"mound",
		"red_mushroom",
		"green_mushroom",
		"blue_mushroom",
--		"mushtree_small", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"mushtree_medium", -- Configurable in the worldgenoverride.lua or already renewable in game
--		"mushtree_tall", -- Configurable in the worldgenoverride.lua or already renewable in game
		"pighouse",
--		"rabbithole", -- Renews itself in the summer
		"rabbithouse",
		"reeds",
		"rock_tough",
		"rock_flintless",
		"slurtlehole",
		"rook",
		"ruins_table",
		"ruins_chair",
		"ruins_vase",
		"ruins_plate",
		"ruins_bowl",
		"ruins_chipbowl",
		"stalagmite",
		"stalagmite_tall",
		"statueharp",
		"statuemaxwell",
		"statueruins",
		"pighead",
		"mermhead",
		"tentacle",
		"ancient_altar",
		"ancient_altar_broken",
		"ruins_statue_mage",
		"ruins_statue_mage_nogem",
		"houndbone",
		"stalagmite_full",
		"stalagmite_med",
		"stalagmite_low",
		"stalagmite_tall_full",
		"stalagmite_tall_med",
		"stalagmite_tall_low",
		"rocky",
		"bat",
		"worm",
		"bishop_nightmare",
		"rook_nightmare",
		"knight_nightmare",
		"resurrectionstone",
		"pandoraschest",
		"houndmound",
		"molehill",
		"pigtorch",
		"spiderden",
		"spiderhole",
		"tallbirdnest",
		"wasphive",
		-- 3rd Party mods
		"appletree",
		"berryblue",
		"berryblu2",
		"berrygree",
		"berrygre2",
		"pineapple",
		"beet_planted",
		"onion_planted",
		"radish_planted",
		"potato_planted",
	}
	for k, prefab in ipairs(renew_normal) do -- Most items
		if customprefabs[prefab] == nil then -- skip custom prefabs
			AddPrefabPostInit(prefab, function(inst)
				if GLOBAL.TheWorld.ismastersim then
					inst:AddComponent("RenewableResourceComponent")
					inst.components.RenewableResourceComponent.delay = math.ceil(WorldResourceRegenerationSpeed) -- Defaults to this
				end
			end)
		end
	end

	local renew_long =	{
		"berrybush_juicy",
		"twiggytree",
		"grassgekko",
		"rock_petrified_tree",
		-- 3rd Party mods
	}
	for k, prefab in ipairs(renew_long) do
		if customprefabs[prefab] == nil then -- skip custom prefabs
			AddPrefabPostInit(prefab, function(inst)
				if GLOBAL.TheWorld.ismastersim then
					inst:AddComponent("RenewableResourceComponent")
					inst.components.RenewableResourceComponent.delay = math.ceil(2*WorldResourceRegenerationSpeed - 1) -- speed*2 would mean today, -1 makes it almost the whole cycle
				end
			end)
		end
	end

	-- custom renew speeds
	for prefab, speed in pairs(customprefabs) do
		if speed > 0 then
			AddPrefabPostInit(prefab, function(inst)
				if GLOBAL.TheWorld.ismastersim then
					inst:AddComponent("RenewableResourceComponent")
					inst.components.RenewableResourceComponent.delay = speed
				end
			end)
		end
	end

	-- Mounds workaround by Muche
	local currentActionDigFn = GLOBAL.ACTIONS.DIG.fn or function() return true end
	GLOBAL.ACTIONS.DIG.fn = function(act) -- Mounds do not disappear, and thus are not marked for respawning normally...
		if act.target and act.target.prefab == "mound" then
--			act.target:DoTaskInTime(TUNING.SEG_TIME, function(inst) -- Old version, but too slow. When server saves in this interval, it breaks the spawn function
			act.target:DoTaskInTime(0.1, function(inst) -- Saving and loading server will cause this task to cancel, thus we don't want a too long time to ensure it will regenerate
				local parent = inst.parent
				if parent ~= nil then
					-- workaround for mound with gravestone as a parent
					-- GetPosition() is 0.35,0,0.35 (i.e. relative to parent), even though GetWorldPosition() is good
					-- Remove() first removes mound from the parent, making GetPosition==GetWorldPosition, i.e. not good
					local x, y, z = inst.Transform:GetWorldPosition()
					inst.Transform:SetPosition(x, y, z)
				end
				inst:Remove()
				if EnableGraveStoneRemoval > 0 then
					if parent ~= nil then
						-- remove parent gravestone as well, so it doesn't spawn its own mound after server reload
						parent:Remove()
					end
				end
			end)
		end
		return currentActionDigFn(act)
	end

	-- Fireflies workaround by Muche
	AddPrefabPostInit("fireflies", function(inst)
		if GLOBAL.TheWorld.ismastersim then
			local oldWorkableFnOnFinishCB = inst.components.workable.onfinish
			inst.components.workable:SetOnFinishCallback(function(inst, worker)
				local RenewableResourceComponent = inst.components.RenewableResourceComponent
				if RenewableResourceComponent and RenewableResourceComponent.shouldAdd then
					GLOBAL.TheWorld.components.ResourceRespawnerComponent:AddInstance(RenewableResourceComponent.inst, RenewableResourceComponent.delay)
				end
				oldWorkableFnOnFinishCB(inst, worker)
			end)
		end
	end)

	-- Turfs Regeneration
	if EnableTurfsRegeneration and GLOBAL.TheNet:GetIsServer() then
		AddComponentPostInit("terraformer", function(self, inst)
			if not GLOBAL.TheWorld.ismastersim then -- redundant due to TheNet:GetIsServer check above, just to be safe
				return
			end
			local old_Terraform = self.Terraform
			self.Terraform = function(self, pt, ...)
				local ret = old_Terraform(self, pt, ...)
				if EnableDebug > 0 then
					print(string.format("[WorldResourceRegeneration|terraformer:Terraform(%s,%s)] inst=%s, ret=%s", tostring(self), tostring(pt), tostring(self.inst), tostring(ret)))
				end

				local world = GLOBAL.TheWorld
				if ret and world and world.components.ResourceRespawnerComponent then
					local x, y = world.Map:GetTileCoordsAtPoint(pt:Get())
					world.components.ResourceRespawnerComponent:AddTurfRegen(x, y)
				end

				return ret
			end
		end)
	end

	local function worldpostinit(inst)
		if GLOBAL.TheWorld.ismastersim then
			inst:AddComponent("ResourceRespawnerComponent")
			inst.components.ResourceRespawnerComponent.spawnFreeZone = SpawnFreeZone
			inst.components.ResourceRespawnerComponent.isdebug = EnableDebug
			inst.components.ResourceRespawnerComponent.restoreTurfs = EnableTurfsRegeneration
			inst.components.ResourceRespawnerComponent:SetRegenSpeed(math.ceil(WorldResourceRegenerationSpeed))
		end
	end

	AddPrefabPostInit("forest", worldpostinit) --Add the Regeneration Module for the Overworld
	AddPrefabPostInit("cave", worldpostinit) --Add the Regeneration Module for the Caves
end

------------------------------------------------------ANNOUNCEMENTS--------------------------------------------------------
local _G = GLOBAL
local TheNet = _G.TheNet
if not(TheNet and TheNet:GetIsServer()) then return end

local STRINGS = _G.STRINGS
local AllPlayers = _G.AllPlayers
local TheNet = _G.TheNet

local announce_open = GetModConfigData("ANNOUNCE_OPEN")
local say_open = GetModConfigData("SAY_OPEN")
local announce_cd = GetModConfigData("ANNOUNCE_CD") * 60
local announce_order = GetModConfigData("ANNOUNCE_ORDER")
local say_showtime = GetModConfigData("SAY_SHOWTIME")
local say_delay = GetModConfigData("SAY_DELAY")
local announce_texts = GetModConfigData("ANNOUNCE_TEXT")
local announce_firstjion_text = GetModConfigData("ANNOUNCE_FIRSTJION_TEXT")
local say_firstjion_text = GetModConfigData("SAY_FIRSTJION_TEXT")
local say_everyjion_text = GetModConfigData("SAY_EVERYJION_TEXT")
local say_death_text = GetModConfigData("SAY_DEATH_TEXT")
local say_resurrection_text = GetModConfigData("SAY_RESURRECTION_TEXT")

if type(announce_texts) ~= "table" then return end
if type(announce_firstjion_text) ~= "string" then return end
if type(say_firstjion_text) ~= "string" then return end
if type(say_everyjion_text) ~= "string" then return end
if type(say_death_text) ~= "string" then return end
if type(say_resurrection_text) ~= "string" then return end

local index = 0
local indexlist = {}
local playerlist = {}

-- Random players:% pds days % pns name, the number of players:% nws world % ncs cave.
-- 随机玩家:% pds 天数 % pns 昵称, 玩家数:% nws 地面 % ncs 洞穴.
local function ReplaceString(str)
	if str == nil or type(str) ~= "string" then
		print("<Advertisement> ReplaceString Error :", str, type(str))
		return ""
	end
	if string.find(str, "%%pns") or string.find(str, "%%pds") or string.find(str, "%%nws") or string.find(str, "%%ncs") then
		local isdedicated = not TheNet:GetServerIsClientHosted()
		local initnum = 0
		local allnum = isdedicated and (#TheNet:GetClientTable() - 1) or #TheNet:GetClientTable()
		local random = math.random(allnum)
		allnum = 0
		for k, v in ipairs(TheNet:GetClientTable()) do
			if not isdedicated or v.performance == nil then
				if v.prefab == "" then initnum = initnum + 1 end
				allnum = allnum + 1
				if random == allnum then
					local charactername = STRINGS.CHARACTER_NAMES[v.prefab] or v.prefab
					str = string.gsub(str, "%%pns", v.name.."("..charactername..")")
					str = string.gsub(str, "%%pds", v.playerage)
				end
			end
		end
		local worldnum = #AllPlayers
		local cavenum = allnum - initnum - worldnum
		if not _G.TheWorld.ismastershard then
			worldnum, cavenum = cavenum, worldnum
		end
		str = string.gsub(str, "%%nws", worldnum)
		str = string.gsub(str, "%%ncs", cavenum)
	end
	return str
end

local function GetAnnounce()
	local message = nil
	if announce_order then
		index = index + 1
		if index > #announce_texts then
			index = index % #announce_texts
		end
		message = announce_texts[index]
	else
		if #indexlist == 0 then
			for k, v in ipairs(announce_texts) do
				table.insert(indexlist, v)
			end
		end
		index = math.random(#indexlist)
		message = indexlist[index]
		table.remove(indexlist, index)
	end
	return ReplaceString(message)
end

local function PublicAnnounce()
	local announce = GetAnnounce()
	if string.len(announce) > 0 then
		TheNet:Announce(announce)
	end
end

AddComponentPostInit("playerspawner", function(OnAnnounce, inst)
	if announce_open and announce_cd > 0 and _G.TheWorld.ismastershard and #announce_texts > 0 then
		_G.scheduler:ExecutePeriodic(announce_cd, PublicAnnounce)
	end

	inst:ListenForEvent("ms_playerjoined", function(inst, player)
		if not player then return end

		if playerlist[player.userid] == nil then
			if announce_open and _G.TheWorld.ismastershard and string.len(announce_firstjion_text) > 0 then
				player:DoTaskInTime(1, function()
					TheNet:Announce(ReplaceString(announce_firstjion_text))
				end)
			end
			if say_open and string.len(say_firstjion_text) > 0 then
				player:DoTaskInTime(say_delay, function(self)
					if self.components and self.components.talker then
						self.components.talker:Say(ReplaceString(say_firstjion_text), say_showtime)
					end
				end)
			end
			playerlist[player.userid] = true
		elseif  say_open and string.len(say_everyjion_text) > 0 then
			player:DoTaskInTime(say_delay, function(self)
				if self.components and self.components.talker then
					self.components.talker:Say(ReplaceString(say_everyjion_text), say_showtime)
				end
			end)
		end

		if say_open then
			if string.len(say_death_text) > 0 then
				player:ListenForEvent("death", function(self)
					self:DoTaskInTime(say_delay, function(self)
						if self.components and self.components.talker then
							self.components.talker:Say(ReplaceString(say_death_text), say_showtime)
						end
					end)
				end)
			end
			if string.len(say_resurrection_text) > 0 then
				player:ListenForEvent("respawnfromghost", function(self)
					self:DoTaskInTime(say_delay, function(self)
						if self.components and self.components.talker then
							self.components.talker:Say(ReplaceString(say_resurrection_text), say_showtime)
						end
					end)
				end)
			end
		end
	end)

end)
