local assets=
{
    Asset("ANIM", "anim/hound_basic.zip"),
    Asset("ANIM", "anim/hound.zip"),
    Asset("ANIM", "anim/hound_ice.zip"),
    Asset("ANIM", "anim/hound_red.zip"),
    Asset("SOUND", "sound/hound.fsb"),
}

local prefabs=
{
    "houndstooth",
    "monstermeat",
}

local brain = require "brains/denguard_brain"

local function retargetfn(inst)
    if inst.components.follower and not inst.components.combat.target then
        local nearest = FindEntity(inst, 100, function(guy)
            return guy:HasTag("skalikas")
        end)
        if nearest and nearest.identity == inst.identity then
            nearest.components.leader:AddFollower(inst)
            inst.components.follower:AddLoyaltyTime(999999)
        elseif nearest then
            inst.components.combat:SetTarget(nearest)
        end
    end
end

local function OnStopFollowing(inst) 
    --inst:RemoveTag("companion") 
    inst.components.health:StartRegen(-5, 0.1)
end

local function OnStartFollowing(inst) 
    --inst:AddTag("companion") 
end

local function KeepTarget(inst, target)
    return inst.components.combat:CanTarget(target) and inst:IsNear(target, 100)
end

local function OnAttacked(inst, data)
    -- They ignore incoming attacks.
end

local function OnNewTarget(inst, data)
--
end

local function OnKilled(inst)
-- You are a horrible person.
end

local function OnSave(inst, data)
    --
end

local function OnPreLoad(inst, data)
    inst.components.health:StartRegen(-5, 0.1)

    if not data then return end

    --
end

local function fn()
    local inst = CreateEntity()
    inst.entity:AddTransform()
    inst.entity:AddAnimState()
    inst.entity:AddSoundEmitter()
    inst.entity:AddDynamicShadow()
    inst.entity:AddNetwork()
    inst.AnimState:SetBank("hound")
    inst.AnimState:SetBuild("hound")
    inst.AnimState:PlayAnimation("idle", true)

    local identity = 0
    inst.identity = 0

  if not TheWorld.ismastersim then
        return inst
    end

    inst.entity:SetPristine()

    inst:AddComponent("locomotor")
    inst.components.locomotor.runspeed = 10
    MakeCharacterPhysics(inst, 10, .5)
    inst.DynamicShadow:SetSize(2.5, 1.5)
    inst.Transform:SetFourFaced()

    inst:SetBrain(brain)
    inst:SetStateGraph("SGhound")

    inst:AddTag("scarytoprey")
    inst:AddTag("notraptrigger")
    inst:AddTag("houndguard")
    --inst:AddComponent("inventoryitem")

    inst:AddComponent("combat")
    inst.components.combat:SetDefaultDamage(20)
    inst.components.combat:SetAttackPeriod(0.5) -- 4x faster!
    inst.components.combat:SetRetargetFunction(3, retargetfn)
    inst.components.combat:SetKeepTargetFunction(KeepTarget)
    inst.components.combat:SetHurtSound("dontstarve/creatures/hound/hurt")
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("attacked", OnAttacked)

    inst:AddComponent("health")
    inst.components.health:SetMaxHealth(1000)
    inst.components.health:StartRegen(-2, 0.1)
    inst.components.health.fire_damage_scale = 1 -- Default

    inst:ListenForEvent("attacked", OnAttacked)
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("death", OnKilled)

    inst:AddComponent("lootdropper")
    inst.components.lootdropper:SetLoot({})

    inst:AddComponent("follower")
    inst:ListenForEvent("stopfollowing", OnStopFollowing)
    inst:ListenForEvent("startfollowing", OnStartFollowing)
    inst.components.follower.maxfollowtime = 9999999
    inst.components.follower:AddLoyaltyTime(999999) 

    inst:AddComponent("inspectable")

    inst.OnSave = OnSave
    inst.OnPreLoad = OnPreLoad

    if math.random() > .8 then
        if math.random() > .5 then
            inst.Transform:SetScale(1.2, 1.2, 1.2)
            inst.AnimState:SetBuild("hound_red")
        else
            inst.Transform:SetScale(0.8, 0.8, 0.8)
            inst.AnimState:SetBuild("hound_ice")
        end
    end

    return inst
end

return Prefab( "common/monsters/denguard", fn, assets, prefabs)