local easing = require("easing")

local assets=
{
    Asset("ANIM", "anim/denmother.zip")
}

local prefabs=
{
}

local brain = require "brains/denmother_brain"

local function CallGuard(inst)
    local x, y, z = inst.Transform:GetWorldPosition()
    local guard = SpawnPrefab("denguard")
    guard.Transform:SetPosition(x, y, z)
    guard.identity = inst.identity

    if inst.components.combat.target then
        guard.components.combat:SetTarget(inst.components.combat.target)
    end
end

local function retargetfn(inst)
    -- There can only be one...
    local nearest = FindEntity(inst, 100, function(guy)
        return guy:HasTag("skalikas")
    end)
    if nearest then
        CallGuard(inst)
        inst.components.combat:SetTarget(nearest)
        inst.components.groundpounder:GroundPound()
    end
end

local function KeepTarget(inst, target)
    return inst.components.combat:CanTarget(target) and inst:IsNear(target, TUNING.HOUND_FOLLOWER_TARGET_KEEP)
end

local function OnAttacked(inst, data)
    inst.components.combat:SetTarget(data.attacker)
end

local function OnNewTarget(inst, data)
--
end

local function OnKilled(inst)
-- You are a horrible person.
end

local function OpenWide(inst)
    inst.AnimState:PlayAnimation("open", true)
    inst.state = "open"
end

local function ShouldAcceptItem(inst, item)
    if item.prefab=="meat" then
        OpenWide(inst)
        return true
    elseif item.prefab=="dragonpie" and math.random() > .75 then
        local x, y, z = inst.Transform:GetWorldPosition()
        SpawnPrefab("denmothercollar").Transform:SetPosition(x, y, z)
        return true
    else
        return false
    end
end

local function OnStartMoon(inst)
    -- Nothing
end

local function OnStartNight(inst)
    -- Nothing
end

local function OnGetItemFromPlayer(inst, giver, item)
    -- Nothing
end

local function OnRefuseItem(inst, item)
    CallGuard(inst)
    -- Ignore Player.
end

local function SetGroundPounderSettings(inst, mode)
    if mode == "normal" then 
        inst.components.groundpounder.damageRings = 0
        inst.components.groundpounder.destructionRings = 0
        inst.components.groundpounder.numRings = 1
        inst.components.groundpounder.groundpounddamagemult = 0
    --[[elseif mode == "hibernation" then 
        inst.components.groundpounder.damageRings = 3
        inst.components.groundpounder.destructionRings = 3
        inst.components.groundpounder.numRings = 4]]
    end
end

local function OnGroundPound(inst)
--
end

local function onupdate(inst, dt)

    -- Have a child at 1% chance every 2 seconds while open.
    if math.random(1,100) > 99 and inst.state == "open" then
        inst.AnimState:PlayAnimation("idle", true)
        inst.state = "idle"
        local x, y, z = inst.Transform:GetWorldPosition()
        SpawnPrefab("lana").Transform:SetPosition(x, y, z)

        inst.components.groundpounder:GroundPound()
    end

    -- There can only be one... (Copy)
    local nearest = FindEntity(inst, 100, function(guy)
        return guy:HasTag("skalikas")
    end)
    if nearest then
        inst.components.combat:SetTarget(nearest)
    end
end

local function OnSave(inst, data)
    -- Nothing
end

local function OnPreLoad(inst, data)
    -- Nothing
end

local function fn()
    local inst = CreateEntity()
    inst.entity:AddTransform()
    inst.entity:AddAnimState()
    inst.entity:AddSoundEmitter()
    inst.entity:AddDynamicShadow()
    inst.entity:AddNetwork()
    inst.AnimState:SetBank("denmother")
    inst.AnimState:SetBuild("denmother")
    inst.AnimState:PlayAnimation("idle", true)

    local state = "idle"
    inst.state = "idle"

    local identity = 0
    inst.identity = math.random()

  if not TheWorld.ismastersim then
        return inst
    end

    inst.entity:SetPristine()

    inst:AddComponent("locomotor")
    inst.components.locomotor.runspeed = 0
    MakeCharacterPhysics(inst, 200, 0.5)
    inst.DynamicShadow:SetSize(2.5, 1.5)
    inst.Transform:SetFourFaced()

    inst:SetBrain(brain)

    inst:AddTag("skalikas")
    inst:AddTag("scarytoprey")
    inst:AddTag("notraptrigger")
    --inst:AddComponent("inventoryitem")

    inst:AddComponent("combat")
    inst.components.combat:SetDefaultDamage(10)
    inst.components.combat:SetAttackPeriod(2)
    inst.components.combat:SetRetargetFunction(10, retargetfn)
    inst.components.combat:SetKeepTargetFunction(KeepTarget)
    inst.components.combat:SetHurtSound("dontstarve/creatures/hound/hurt")
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("attacked", OnAttacked)

    inst:AddComponent("health")
    inst.components.health:SetMaxHealth(1000)
    inst.components.health:StartRegen(1, 0.2) -- 5hp/sec
    inst.components.health.fire_damage_scale = 0 -- None

    inst:ListenForEvent("attacked", OnAttacked)
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("death", OnKilled)

    inst:AddComponent("inspectable")

    inst:AddComponent("trader")
    inst.components.trader:SetAcceptTest(ShouldAcceptItem)
    inst.components.trader.onaccept = OnGetItemFromPlayer
    inst.components.trader.onrefuse = OnRefuseItem

    inst:AddComponent("sanityaura")
    inst.components.sanityaura.aura = 0

    inst:WatchWorldState("startnight", OnStartNight)
    inst:WatchWorldState("isfullmoon", OnStartMoon)

    inst:AddComponent("leader")

    inst:AddComponent("groundpounder")
    inst.components.groundpounder.destroyer = false
    SetGroundPounderSettings(inst, "normal")

    inst:DoPeriodicTask(2, onupdate, nil, 1) --arguments:(time, fn, delay, fnArguments)

    inst.OnSave = OnSave
    inst.OnPreLoad = OnPreLoad

    return inst
end

return Prefab( "common/monsters/denmother", fn, assets, prefabs)
