local easing = require("easing")

local assets=
{
    Asset("ANIM", "anim/hound_basic.zip"),
    Asset("ANIM", "anim/hound.zip"),
    Asset("ANIM", "anim/hound_ice.zip"),
    Asset("ANIM", "anim/hound_red.zip"),
    Asset("ANIM", "anim/hound_gold.zip"),
    Asset("ANIM", "anim/hound_nightmare.zip"),
    Asset("SOUND", "sound/hound.fsb"),
}

local prefabs=
{
    "houndstooth",
    "monstermeat",
}

local brain = require "brains/lana_brain"

local WAKE_TO_FOLLOW_DISTANCE = 6
local Names = {"Ladda","Lady","Lambda","Lana","Lapis","Lazarus","Lazy","Lela","Lemonade","Leroy","Lex","Lil' Lucifer","Lilah","Lilith","Lilly","Linn","Logsworth","Loki","Lolly","Lolo","Loo","Loony","Lord","Loser","Love","Lucile","Lucy","Lula","Lulu","Lulu","Luna"}
local BonusNames = {}
-- Names list is 1-31
-- Bonus list is 1-8

local function retargetfn(inst)
    if inst.components.follower and not inst.components.follower.leader then
        local nearest = FindEntity(inst, 100, function(guy)
            return guy:HasTag("mamahound")
        end)
        if nearest and nearest.components.leader then
            nearest.components.leader:AddFollower(inst)
            inst.components.follower:AddLoyaltyTime(999999) 
        end
    end
end

local function  DistractedHound(inst)
    -- The younger a hound is, the more likely it is to be distracted by birds!
    -- 19 day old puppies are never distracted.
    local randomizer = math.random()
    if randomizer > (.025*inst.age)+.525 then
        local nearbird = FindEntity(inst,25, function (birb)
                return birb:HasTag("bird")
            end)
        if nearbird then
            inst.components.combat:SetTarget(nearbird)
        end
    end
end

local function OnStopFollowing(inst) 
    --inst:RemoveTag("companion") 
end

local function OnStartFollowing(inst) 
    --inst:AddTag("companion") 
end

local function KeepTarget(inst, target)
    return inst.components.combat:CanTarget(target) and inst:IsNear(target, TUNING.HOUND_FOLLOWER_TARGET_KEEP)
end

local function OnAttacked(inst, data)
    inst.components.combat:SetTarget(data.attacker)
end

local function OnNewTarget(inst, data)
--
end

local function OnKilled(inst)
-- You are a horrible person.
end

local function SpawnAdult(inst)
    --print("lana - SpawnAdult")

    local koko

    -- Change into the koko variant based on the puppy variant.
    if inst.flavor == "normal" then
        koko = SpawnPrefab("koko")
    elseif inst.flavor == "hotlana" then
        koko = SpawnPrefab("kokoflame")
    elseif inst.flavor == "coldlana" then
        koko = SpawnPrefab("kokofrost")
    elseif inst.flavor == "crazylana" then
        koko = SpawnPrefab("kokocrazy")
    elseif inst.flavor == "shinylana" then
        koko = SpawnPrefab("kokoshiny")
    else
        koko = SpawnPrefab("koko")
    end
    koko.Transform:SetPosition(inst.Transform:GetWorldPosition())
    koko.sg:GoToState("idle")
    koko.components.combat:BlankOutAttacks(6 + math.random()*4)
    koko.MyName = inst.MyName
    koko.components.named:SetName(inst.MyName)
    koko.components.inspectable.description = ("Hello, " .. inst.MyName)

    inst:Remove()
end

local function icechange(inst)
    inst.AnimState:SetBuild("hound_ice")

    inst.components.combat:SetDefaultDamage(17) -- Normal=20 Special=30
    inst.components.combat:SetAttackPeriod(2) -- Normal Speed
    inst.components.locomotor.runspeed = 10
    inst.components.health:SetMaxHealth(750)
    inst.components.health:StartRegen(15, 0.2) -- 3X as strong
    inst.components.health.fire_damage_scale = 0.8 -- Fire resistant
    inst.flavor = "coldlana"
end

local function firechange(inst)
    inst.AnimState:SetBuild("hound_red")

    inst.components.combat:SetDefaultDamage(15) -- Normal=20 Special=30
    inst.components.combat:SetAttackPeriod(1.5) -- Faster Speed (4:3 attacks=120:105 dps) (60 base)
    inst.components.locomotor.runspeed = 10
    inst.components.health:SetMaxHealth(750)
    inst.components.health:StartRegen(10, 0.2) -- 2X as strong
    inst.components.health.fire_damage_scale = 0 -- Fireproof
    inst.flavor = "hotlana"
end

local function crazychange(inst)
    inst.AnimState:SetBuild("hound_nightmare")

    inst.components.combat:SetDefaultDamage(20) -- Normal=20 Special=30
    inst.components.combat:SetAttackPeriod(1.5) -- Faster Speed (4:3 attacks=120:105 dps) (60 base)
    inst.components.locomotor.runspeed = 10
    inst.components.health:SetMaxHealth(600)
    inst.components.health:StartRegen(7, 0.1) -- 2X as fast, a little stronger
    inst.components.health.fire_damage_scale = 0 -- Fireproof
    inst.flavor = "crazylana"

    inst.components.sanityaura.aura = -TUNING.SANITYAURA_MED
end

local function goldchange(inst)
    inst.AnimState:SetBuild("hound_gold")

    inst.components.combat:SetDefaultDamage(10) -- Normal=20 Special=30
    inst.components.combat:SetAttackPeriod(0.5) -- Faster Speed (4:1 attacks=40 dps)
    inst.components.locomotor.runspeed = 10
    inst.components.health:SetMaxHealth(1000 + inst.age)
    inst.components.health:StartRegen(10, 0.1) -- 2X strength/speed (Net 4X)
    inst.components.health.fire_damage_scale = 0.5 -- Half Damage
    inst.flavor = "shinylana"
end

local function OnStartMoon(inst)
    if inst.age >= 25 then
        inst.AnimState:SetBuild("hound_gold")

        inst.components.combat:SetDefaultDamage(10) -- Normal=20 Special=30
        inst.components.combat:SetAttackPeriod(0.5) -- Faster Speed (4:1 attacks=40 dps)
        inst.components.locomotor.runspeed = 10
        inst.components.health:SetMaxHealth(1000)
        inst.components.health:StartRegen(10, 0.1) -- 2X strength/speed (Net 4X)
        inst.components.health.fire_damage_scale = 0.5 -- Half Damage
        inst.flavor = "shinylana"
    end
end

local function OnStartNight(inst)
    -- TheWorld.net.components.clock:OnUpdate(30*16*1) To Skip Days
    -- Puppies take 19 days to grow up.
    if inst.age < 19 then
        inst.age = inst.age+1
    
        local scale = easing.inQuad(math.random(), .5+(inst.age*.025), .025, 1)
        inst.Transform:SetScale(scale, scale, scale)
    else
        SpawnAdult(inst)
    end

    if inst.flavor == "normal" then
        if inst.age <= 50 then
            inst.components.health:SetMaxHealth(500+(inst.age*10)) -- Default hounds go from 1K-2K health in 50 days.
        else
            inst.components.health:SetMaxHealth(1000) -- Cap is 2K, can turn gold at 1.5K to instantly go to 2K.
        end
    end
end

local function ShouldAcceptItem(inst, item)

    local nearest = FindEntity(inst, 100, function(guy)
    return guy:HasTag("player")
    end)

    if nearest == nil then
        return false
    end

    -- Feeding puppies currently disabled.
    if  nearest:HasTag("Murderer!") ~= true and nearest.userid == "KU_M27Zfz_d" then
        if item.prefab == "bluegem" and not inst.flavor == "shinylana" then
            if inst.flavor == "coldlana" then
                return false
            else
                icechange(inst)
                return true
            end

        elseif item.prefab == "redgem" and not inst.flavor == "shinylana" then
            if inst.flavor == "hotlana" then
                return false
            else
                firechange(inst)
                return true
            end

        elseif item.prefab == "nightmaretreat" and not inst.flavor == "shinylana" then
            if inst.flavor == "crazylana" then
                return false
            else
                if inst.crazy <= 1 then
                    inst.crazy = inst.crazy+1
                end
                if inst.crazy == 2 then
                    crazychange(inst)
                end
                return true
            end
        elseif item.prefab == "bonestew" then
            OnStartNight(inst)
            OnStartNight(inst)
            return true
        end
            --Debug Only
        if nearest.userid == "KU_M27Zfz_d" then
            if item.prefab == "goldnugget" then
                if inst.flavor == "normal" then
                    goldchange(inst)
                    return true
                end
            elseif item.prefab == "seeds" then
                OnStartNight(inst)
                return true
            elseif item.prefab == "dragonpie" then
                inst.age = inst.age + 20
                OnStartNight(inst)
                return true
            end 
        else
            return false
        end
    else
        return false
    end
end

local function OnGetItemFromPlayer(inst, giver, item)

end

local function OnRefuseItem(inst, item)
    -- Ignore Player.
end

local function ShouldWakeUp(inst)
    if inst.components.health.currenthealth < (inst.components.health.maxhealth) then
        return false
    else
        return ((inst.components.follower and inst.components.follower.leader and not inst.components.follower:IsNearLeader(WAKE_TO_FOLLOW_DISTANCE)) or (inst.components.combat and inst.components.combat.target) or (TheWorld.state.isnight == false))
    end
end

local function ShouldSleep(inst)
    if inst.components.health.currenthealth < (inst.components.health.maxhealth/2) then
        return true
    else
        return TheWorld.state.isnight and not (inst.components.combat and inst.components.combat.target) and not (inst.components.follower and inst.components.follower.leader and not inst.components.follower:IsNearLeader(WAKE_TO_FOLLOW_DISTANCE))
    end
end

local function onupdate(inst, dt)
    -- Every three seconds, test if the puppy gets distracted by birds.
    DistractedHound(inst)
end

local function OnSave(inst, data)
    data.flavor = inst.flavor
    data.age = inst.age
    data.crazy = inst.crazy
    data.MyName = inst.MyName
end

local function OnPreLoad(inst, data)
    if not data then return end
    if data.flavor == "coldlana" then
        icechange(inst)
    elseif data.flavor == "hotlana" then
        firechange(inst)
    elseif data.flavor == "shinylana" then
        goldchange(inst)
    elseif data.flavor == "crazylana" then
        crazychange(inst)
    end
    
    inst.age = data.age

    local scale = easing.inQuad(math.random(), .5+(inst.age*.025), .025, 1) -- Base, Variance, ???
    inst.Transform:SetScale(scale, scale, scale)

    if inst.flavor == "normal" then
        if inst.age <= 1000 then
            inst.components.health:SetMaxHealth(500+(inst.age*10)) -- Default hounds go from 1K-2K health in 50 days.
        else
            inst.components.health:SetMaxHealth(1000) -- Cap is 2K, can turn gold at 1.5K to instantly go to 2K.
        end
    end

    inst.crazy = data.crazy

    inst.MyName = data.MyName
    if data.MyName ~= nil then
        inst.components.inspectable.description = (data.MyName .. " is so adorable!")
    end
end

local function fn()
    local inst = CreateEntity()
    inst.entity:AddTransform()
    inst.entity:AddAnimState()
    inst.entity:AddSoundEmitter()
    inst.entity:AddDynamicShadow()
    inst.entity:AddNetwork()
    inst.AnimState:SetBank("hound")
    inst.AnimState:SetBuild("hound")
    inst.AnimState:PlayAnimation("idle", true)

    local flavor = "normal"
    inst.flavor = "normal"

    local age = 0
    inst.age = 0

    local crazy = 0
    inst.crazy = 0

    local scale = easing.inQuad(math.random(), .5+(inst.age*.025), .025, 1) -- Base, Variance, ???
    inst.Transform:SetScale(scale, scale, scale)
    -- .5+(inst.age*.02), .1 // 20 days to adult, short-varied, faster growth
    -- .4+(inst.age*.025), .1 // 20 days to adult, multi-varied, slower growth (smaller min size)
    -- .5+(inst.age*.025), .05 // 18 days to adult, long-varied, decent growth
    -- .5+(inst.age*.01), .05 // 45 days to adult, long-varied, unstable growth
    -- .5+(inst.age*.025), .025 // 19 days to adult, very long-varied, consistant growth

    if not TheWorld.ismastersim then
        return inst
    end

    inst.entity:SetPristine()

    inst:AddComponent("locomotor")
    inst.components.locomotor.runspeed = 10
    MakeCharacterPhysics(inst, 10, .5)
    inst.DynamicShadow:SetSize(2.5, 1.5)
    inst.Transform:SetFourFaced()

    inst:SetBrain(brain)
    inst:SetStateGraph("SGhound")

    inst:AddTag("companion")
    inst:AddTag("scarytoprey")
    inst:AddTag("notraptrigger")
    --inst:AddComponent("inventoryitem")

    inst:AddComponent("combat")
    inst.components.combat:SetDefaultDamage(10)
    inst.components.combat:SetAttackPeriod(2)
    inst.components.combat:SetRetargetFunction(3, retargetfn)
    inst.components.combat:SetKeepTargetFunction(KeepTarget)
    inst.components.combat:SetHurtSound("dontstarve/creatures/hound/hurt")
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("attacked", OnAttacked)

    inst:AddComponent("health")
    inst.components.health:SetMaxHealth(500)
    inst.components.health:StartRegen(5, 0.2)
    inst.components.health.fire_damage_scale = 1 -- Default

    inst:ListenForEvent("attacked", OnAttacked)
    inst:ListenForEvent("newcombattarget", OnNewTarget)
    inst:ListenForEvent("death", OnKilled)

    inst:AddComponent("lootdropper")
    inst.components.lootdropper:SetLoot({"houndstooth"})

    inst:AddComponent("follower")
    inst:ListenForEvent("stopfollowing", OnStopFollowing)
    inst:ListenForEvent("startfollowing", OnStartFollowing)
    inst.components.follower.maxfollowtime = 9999999
    inst.components.follower.keepdeadleader = true

    inst:AddComponent("trader")
    inst.components.trader:SetAcceptTest(ShouldAcceptItem)
    inst.components.trader.onaccept = OnGetItemFromPlayer
    inst.components.trader.onrefuse = OnRefuseItem

    inst:AddComponent("sleeper")
    inst.components.sleeper:SetResistance(3)
    inst.components.sleeper.testperiod = GetRandomWithVariance(3, 2)
    inst.components.sleeper:SetSleepTest(ShouldSleep)
    inst.components.sleeper:SetWakeTest(ShouldWakeUp)

    inst:AddComponent("sanityaura")
    inst.components.sanityaura.aura = 0

    inst:WatchWorldState("startnight", OnStartNight)
    inst:WatchWorldState("isfullmoon", OnStartMoon)

    -- Give then a name unless they already have one!
    if inst.MyName == nil then
        if math.random(1,100)>99 then
            inst.MyName = BonusNames[math.random(1, 8)] -- Rare names, 1% chance per hound.
        else
            inst.MyName = Names[math.random(1, 31)]
        end
    end

    inst:AddComponent("named")
    inst.components.named:SetName(inst.MyName)

    inst:AddComponent("inspectable")
    inst.components.inspectable.description = (inst.MyName .. " is so adorable!")

    inst:DoPeriodicTask(3, onupdate, nil, 1) --arguments:(time, fn, delay, fnArguments)

    inst.OnSave = OnSave
    inst.OnPreLoad = OnPreLoad

    return inst
end

return Prefab( "common/monsters/lana", fn, assets, prefabs)
